from typing import Optional, List, Union

import torch
from torch import nn, Tensor
import torch.nn.functional as F
from torch.nn import Parameter, Sequential, Linear, ReLU, Dropout
from torch.nn import BatchNorm1d, LayerNorm, InstanceNorm1d

import torch_sparse
from torch_sparse import SparseTensor, matmul, fill_diag, sum as sparsesum, mul
from torch_sparse import SparseTensor
from torch_scatter import scatter, scatter_softmax
from torch_scatter import scatter
from torch_scatter import scatter_add

from torch_geometric.nn import MessagePassing, DeepGCNLayer
from torch_geometric.nn.norm import MessageNorm
from torch_geometric.utils import add_self_loops, degree, softmax
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool
from torch_geometric.utils.num_nodes import maybe_num_nodes

from auglichem.utils import NUM_ATOM_TYPE, NUM_CHIRALITY_TAG, NUM_BOND_TYPE, NUM_BOND_DIRECTION


class MLP(Sequential):
    def __init__(self, channels: List[int], norm: Optional[str] = None,
                 bias: bool = True, dropout: float = 0.):
        m = []
        for i in range(1, len(channels)):
            m.append(Linear(channels[i - 1], channels[i], bias))

            if i < len(channels) - 1:
                if norm and norm == 'batch':
                    m.append(BatchNorm1d(channels[i], affine=True))
                elif norm and norm == 'layer':
                    m.append(LayerNorm(channels[i], elementwise_affine=True))
                elif norm and norm == 'instance':
                    m.append(InstanceNorm1d(channels[i], affine=False))
                elif norm:
                    raise NotImplementedError(
                        f'Normalization layer "{norm}" not supported.')
                m.append(ReLU())
                m.append(Dropout(dropout))

        super(MLP, self).__init__(*m)


class GENConv(MessagePassing):
    def __init__(self, emb_dim: int, 
                 aggr: str = 'softmax', t: float = 1.0, learn_t: bool = False,
                 p: float = 1.0, learn_p: bool = False, msg_norm: bool = False,
                 learn_msg_scale: bool = False, norm: str = 'batch',
                 num_layer: int = 2, eps: float = 1e-7):
        
        super(GENConv, self).__init__()

        self.emb_dim = emb_dim
        self.aggr = aggr
        self.eps = eps

        assert aggr in ['softmax', 'softmax_sg', 'power', 'add', 'mean', 'max']

        channels = [emb_dim]
        for i in range(num_layer - 1):
            channels.append(emb_dim * 2)
        channels.append(emb_dim)
        self.mlp = MLP(channels, norm=norm)

        self.msg_norm = MessageNorm(learn_msg_scale) if msg_norm else None

        self.initial_t = t
        self.initial_p = p

        if learn_t and aggr == 'softmax':
            self.t = Parameter(torch.Tensor([t]), requires_grad=True)
        else:
            self.t = t

        if learn_p:
            self.p = Parameter(torch.Tensor([p]), requires_grad=True)
        else:
            self.p = p
        
        self.edge_embedding1 = nn.Embedding(NUM_BOND_TYPE, emb_dim)
        self.edge_embedding2 = nn.Embedding(NUM_BOND_DIRECTION, emb_dim)

        nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        nn.init.xavier_uniform_(self.edge_embedding2.weight.data)

    def reset_parameters(self):
        reset(self.mlp)
        if self.msg_norm is not None:
            self.msg_norm.reset_parameters()
        if self.t and isinstance(self.t, Tensor):
            self.t.data.fill_(self.initial_t)
        if self.p and isinstance(self.p, Tensor):
            self.p.data.fill_(self.initial_p)

    def forward(self, x, edge_index, edge_attr, size=None):

        edge_index = add_self_loops(edge_index, num_nodes=x.size(0))[0]
        # print(edge_index)

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:,0] = 4 #bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        edge_embeddings = self.edge_embedding1(edge_attr[:,0]) + self.edge_embedding2(edge_attr[:,1])

        # propagate_type: (x: OptPairTensor, edge_attr: OptTensor)
        out = self.propagate(edge_index, x=x, edge_attr=edge_embeddings, size=size)

        if self.msg_norm is not None:
            out = self.msg_norm(x[0], out)

        x_r = x[1]
        if x_r is not None:
            out += x_r

        return self.mlp(out)

    def message(self, x_j, edge_attr):
        msg = x_j if edge_attr is None else x_j + edge_attr
        return F.relu(msg) + self.eps

    def aggregate(self, inputs: Tensor, index: Tensor,
                  dim_size: Optional[int] = None) -> Tensor:

        if self.aggr == 'softmax':
            out = scatter_softmax(inputs * self.t, index, dim=self.node_dim)
            return scatter(inputs * out, index, dim=self.node_dim,
                           dim_size=dim_size, reduce='sum')

        elif self.aggr == 'softmax_sg':
            out = scatter_softmax(inputs * self.t, index,
                                  dim=self.node_dim).detach()
            return scatter(inputs * out, index, dim=self.node_dim,
                           dim_size=dim_size, reduce='sum')

        elif self.aggr == 'power':
            min_value, max_value = 1e-7, 1e1
            torch.clamp_(inputs, min_value, max_value)
            out = scatter(torch.pow(inputs, self.p), index, dim=self.node_dim,
                          dim_size=dim_size, reduce='mean')
            torch.clamp_(out, min_value, max_value)
            return torch.pow(out, 1 / self.p)

        else:
            return scatter(inputs, index, dim=self.node_dim, dim_size=dim_size,
                           reduce=self.aggr)


class DeepGCN(nn.Module):
    def __init__(self, 
        task='classification', num_layer=28, emb_dim=300, 
        feat_dim=256, pool='mean', drop_ratio=0, output_dim=None, **kwargs
    ):
        """
          Inputs:
          -------
          emb_dim (int): Edge feature dimensionality.
          aggr (str, optional, default='softmax'): Aggregate function, one of 'softmax',
                                  'softmax_sg', 'power', 'add', 'mean', 'max'.
          t (float optional, default=1.0): Scaling parameter for softmax and
                                  softmax_sg aggregation.
          learn_t (bool optional, default=False): Flag to learn t or not.
          p (float, optional, default=1.0): Power used for power aggreagation.
          learn_p (bool, optional, default=False): Flag to learn p or not.
          msg_norm (bool, optional, default=False): Flag to normalize messages or not.
          learn_msg_scale (bool, optional, default=False): Flag to learn message norm or not.
          norm (str, optional, default ='batch'): Type of norm to use in MLP.
                                  One of 'batch', 'layer',     or 'instance'.
          num_layer (int, optional, default=2): Number of layers in the network.
          eps (float, optional, default=1e-7): Small value to add to message output.

          Outputs:
          --------
          None
        """
        super(DeepGCN, self).__init__()
        self.num_layer = num_layer
        self.emb_dim = emb_dim
        self.feat_dim = feat_dim
        self.drop_ratio = drop_ratio
        self.task = task
        if(output_dim == None):
            self.output_dim = 1
        else:
            self.output_dim = output_dim

        self.x_embedding1 = nn.Embedding(NUM_ATOM_TYPE, emb_dim)
        self.x_embedding2 = nn.Embedding(NUM_CHIRALITY_TAG, emb_dim)

        nn.init.xavier_uniform_(self.x_embedding1.weight.data)
        nn.init.xavier_uniform_(self.x_embedding2.weight.data)

        # List of MLPs
        self.gnns = torch.nn.ModuleList()
        for i in range(1, num_layer + 1):
            conv = GENConv(emb_dim, aggr='softmax',
                           t=1.0, learn_t=True, num_layer=2, norm='layer')
            norm = LayerNorm(emb_dim, elementwise_affine=True)
            act = ReLU(inplace=True)

            layer = DeepGCNLayer(conv, norm, act, block='res+', dropout=0.1,
                                 ckpt_grad=i % 3)
            self.gnns.append(layer)

        if pool == 'mean':
            self.pool = global_mean_pool
        elif pool == 'add':
            self.pool = global_add_pool
        elif pool == 'max':
            self.pool = global_max_pool
        else:
            raise ValueError('Not defined pooling!')
        
        self.feat_lin = nn.Linear(self.emb_dim, self.feat_dim)

        if self.task == 'classification':
            self.output_dim *= 2
            self.pred_lin = nn.Sequential(
                nn.Linear(self.feat_dim, self.feat_dim//2), 
                nn.ReLU(inplace=True),
                nn.Linear(self.feat_dim//2, self.output_dim)
            )
        elif self.task == 'regression':
            self.pred_lin = nn.Sequential(
                nn.Linear(self.feat_dim, self.feat_dim//2), 
                nn.ReLU(inplace=True),
                nn.Linear(self.feat_dim//2, self.feat_dim//2),
                nn.ReLU(inplace=True),
                nn.Linear(self.feat_dim//2, self.feat_dim//4),
                nn.ReLU(inplace=True),
                nn.Linear(self.feat_dim//4, self.output_dim)
            )

    def forward(self, data):
        x = data.x
        edge_index = data.edge_index
        edge_attr = data.edge_attr

        h = self.x_embedding1(x[:,0]) + self.x_embedding2(x[:,1])

        h = self.gnns[0].conv(h, edge_index, edge_attr)
        for layer in self.gnns[1:]:
            h = layer(h, edge_index, edge_attr)
            h = F.dropout(h, self.drop_ratio, training=self.training)
        h = self.gnns[0].act(self.gnns[0].norm(h))

        h = self.feat_lin(h)
        h = self.pool(h, data.batch)

        return h, self.pred_lin(h)

    def load_my_state_dict(self, state_dict):
        own_state = self.state_dict()
        for name, param in state_dict.items():
            if name not in own_state:
                continue
            if isinstance(param, nn.parameter.Parameter):
                # backwards compatibility for serialized parameters
                param = param.data
            own_state[name].copy_(param)



from typing import Optional

import os
import warnings
import os.path as osp
from math import pi as PI

import torch
import torch.nn.functional as F
from torch.nn import Embedding, Sequential, Linear, ModuleList
import numpy as np

from torch_scatter import scatter
from torch_geometric.data.makedirs import makedirs
from torch_geometric.data import download_url, extract_zip, Dataset
from torch_geometric.nn import MessagePassing
from ._radius_graph import radius_graph

qm9_target_dict = {
    0: 'dipole_moment',
    1: 'isotropic_polarizability',
    2: 'homo',
    3: 'lumo',
    4: 'gap',
    5: 'electronic_spatial_extent',
    6: 'zpve',
    7: 'energy_U0',
    8: 'energy_U',
    9: 'enthalpy_H',
    10: 'free_energy',
    11: 'heat_capacity',
}


class SchNet(torch.nn.Module):
    r"""The continuous-filter convolutional neural network SchNet from the
    `"SchNet: A Continuous-filter Convolutional Neural Network for Modeling
    Quantum Interactions" <https://arxiv.org/abs/1706.08566>`_ paper that uses
    the interactions blocks of the form

    .. math::
        \mathbf{x}^{\prime}_i = \sum_{j \in \mathcal{N}(i)} \mathbf{x}_j \odot
        h_{\mathbf{\Theta}} ( \exp(-\gamma(\mathbf{e}_{j,i} - \mathbf{\mu}))),

    here :math:`h_{\mathbf{\Theta}}` denotes an MLP and
    :math:`\mathbf{e}_{j,i}` denotes the interatomic distances between atoms.

    .. note::

        For an example of using a pretrained SchNet variant, see
        `examples/qm9_pretrained_schnet.py
        <https://github.com/rusty1s/pytorch_geometric/blob/master/examples/
        qm9_pretrained_schnet.py>`_.

    Args:
        hidden_channels (int, optional): Hidden embedding size.
            (default: :obj:`128`)
        num_filters (int, optional): The number of filters to use.
            (default: :obj:`128`)
        num_interactions (int, optional): The number of interaction blocks.
            (default: :obj:`6`)
        num_gaussians (int, optional): The number of gaussians :math:`\mu`.
            (default: :obj:`50`)
        cutoff (float, optional): Cutoff distance for interatomic interactions.
            (default: :obj:`10.0`)
        max_num_neighbors (int, optional): The maximum number of neighbors to
            collect for each node within the :attr:`cutoff` distance.
            (default: :obj:`32`)
        readout (string, optional): Whether to apply :obj:`"add"` or
            :obj:`"mean"` global aggregation. (default: :obj:`"add"`)
        dipole (bool, optional): If set to :obj:`True`, will use the magnitude
            of the dipole moment to make the final prediction, *e.g.*, for
            target 0 of :class:`torch_geometric.datasets.QM9`.
            (default: :obj:`False`)
        mean (float, optional): The mean of the property to predict.
            (default: :obj:`None`)
        std (float, optional): The standard deviation of the property to
            predict. (default: :obj:`None`)
        atomref (torch.Tensor, optional): The reference of single-atom
            properties.
            Expects a vector of shape :obj:`(max_atomic_number, )`.
    """

    url = 'http://www.quantum-machine.org/datasets/trained_schnet_models.zip'

    def __init__(self, hidden_channels: int = 128, num_filters: int = 128,
                 num_interactions: int = 6, num_gaussians: int = 50,
                 cutoff: float = 10.0, max_num_neighbors: int = 32,
                 readout: str = 'add', dipole: bool = False,
                 mean: Optional[float] = None, std: Optional[float] = None,
                 atomref: Optional[torch.Tensor] = None):
        super(SchNet, self).__init__()

        import ase

        self.hidden_channels = hidden_channels
        self.num_filters = num_filters
        self.num_interactions = num_interactions
        self.num_gaussians = num_gaussians
        self.cutoff = cutoff
        self.max_num_neighbors = max_num_neighbors
        self.readout = readout
        self.dipole = dipole
        self.readout = 'add' if self.dipole else self.readout
        self.mean = mean
        self.std = std
        self.scale = None

        atomic_mass = torch.from_numpy(ase.data.atomic_masses)
        self.register_buffer('atomic_mass', atomic_mass)

        self.embedding = Embedding(100, hidden_channels)
        self.distance_expansion = GaussianSmearing(0.0, cutoff, num_gaussians)

        self.interactions = ModuleList()
        for _ in range(num_interactions):
            block = InteractionBlock(hidden_channels, num_gaussians,
                                     num_filters, cutoff)
            self.interactions.append(block)

        self.lin1 = Linear(hidden_channels, hidden_channels // 2)
        self.act = ShiftedSoftplus()
        self.lin2 = Linear(hidden_channels // 2, 1)

        self.register_buffer('initial_atomref', atomref)
        self.atomref = None
        if atomref is not None:
            self.atomref = Embedding(100, 1)
            self.atomref.weight.data.copy_(atomref)

        self.reset_parameters()

    def reset_parameters(self):
        self.embedding.reset_parameters()
        for interaction in self.interactions:
            interaction.reset_parameters()
        torch.nn.init.xavier_uniform_(self.lin1.weight)
        self.lin1.bias.data.fill_(0)
        torch.nn.init.xavier_uniform_(self.lin2.weight)
        self.lin2.bias.data.fill_(0)
        if self.atomref is not None:
            self.atomref.weight.data.copy_(self.initial_atomref)


    def forward(self, data, batch=None):
        """"""
        z = data.atomics
        pos = data.pos
        if(batch is None): # If we're using the CrystalDatasetWrapper loaders, data has batch
            try:
                batch = data.batch
            except AttributeError:
                pass
        assert z.dim() == 1 and z.dtype == torch.long
        batch = torch.zeros_like(z) if batch is None else batch

        h = self.embedding(z)

        edge_index = radius_graph(pos, r=self.cutoff, batch=batch,
                                  max_num_neighbors=self.max_num_neighbors)
        row, col = edge_index
        edge_weight = (pos[row] - pos[col]).norm(dim=-1)
        edge_attr = self.distance_expansion(edge_weight)

        for interaction in self.interactions:
            h = h + interaction(h, edge_index, edge_weight, edge_attr)

        h = self.lin1(h)
        h = self.act(h)
        h = self.lin2(h)

        if self.dipole:
            # Get center of mass.
            mass = self.atomic_mass[z].view(-1, 1)
            c = scatter(mass * pos, batch, dim=0) / scatter(mass, batch, dim=0)
            h = h * (pos - c[batch])

        if not self.dipole and self.mean is not None and self.std is not None:
            h = h * self.std + self.mean

        if not self.dipole and self.atomref is not None:
            h = h + self.atomref(z)

        out = scatter(h, batch, dim=0, reduce=self.readout)

        if self.dipole:
            out = torch.norm(out, dim=-1, keepdim=True)

        if self.scale is not None:
            out = self.scale * out

        return out


    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'hidden_channels={self.hidden_channels}, '
                f'num_filters={self.num_filters}, '
                f'num_interactions={self.num_interactions}, '
                f'num_gaussians={self.num_gaussians}, '
                f'cutoff={self.cutoff})')



class InteractionBlock(torch.nn.Module):
    def __init__(self, hidden_channels, num_gaussians, num_filters, cutoff):
        super(InteractionBlock, self).__init__()
        self.mlp = Sequential(
            Linear(num_gaussians, num_filters),
            ShiftedSoftplus(),
            Linear(num_filters, num_filters),
        )
        self.conv = CFConv(hidden_channels, hidden_channels, num_filters,
                           self.mlp, cutoff)
        self.act = ShiftedSoftplus()
        self.lin = Linear(hidden_channels, hidden_channels)

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.mlp[0].weight)
        self.mlp[0].bias.data.fill_(0)
        torch.nn.init.xavier_uniform_(self.mlp[2].weight)
        self.mlp[0].bias.data.fill_(0)
        self.conv.reset_parameters()
        torch.nn.init.xavier_uniform_(self.lin.weight)
        self.lin.bias.data.fill_(0)

    def forward(self, x, edge_index, edge_weight, edge_attr):
        x = self.conv(x, edge_index, edge_weight, edge_attr)
        x = self.act(x)
        x = self.lin(x)
        return x


class CFConv(MessagePassing):
    def __init__(self, in_channels, out_channels, num_filters, nn, cutoff):
        super(CFConv, self).__init__(aggr='add')
        self.lin1 = Linear(in_channels, num_filters, bias=False)
        self.lin2 = Linear(num_filters, out_channels)
        self.nn = nn
        self.cutoff = cutoff

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.lin1.weight)
        torch.nn.init.xavier_uniform_(self.lin2.weight)
        self.lin2.bias.data.fill_(0)

    def forward(self, x, edge_index, edge_weight, edge_attr):
        C = 0.5 * (torch.cos(edge_weight * PI / self.cutoff) + 1.0)
        W = self.nn(edge_attr) * C.view(-1, 1)

        x = self.lin1(x)
        x = self.propagate(edge_index, x=x, W=W)
        x = self.lin2(x)
        return x

    def message(self, x_j, W):
        return x_j * W


class GaussianSmearing(torch.nn.Module):
    def __init__(self, start=0.0, stop= 5.0, num_gaussians=50):
        super(GaussianSmearing, self).__init__()
        offset = torch.linspace(0.0, 5.0, num_gaussians)
        self.coeff = -0.5 / (offset[1] - offset[0]).item()**2
        self.register_buffer('offset', offset)

    def forward(self, dist):
        dist = dist.view(-1, 1) - self.offset.view(1, -1)
        return torch.exp(self.coeff * torch.pow(dist, 2))


class ShiftedSoftplus(torch.nn.Module):
    def __init__(self):
        super(ShiftedSoftplus, self).__init__()
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, x):
        return F.softplus(x) - self.shift


if __name__ == "__main__":
    model = SchNet()
    print(model)



import os
import csv
import math
import random
import numpy as np
from copy import deepcopy
from operator import itemgetter
import warnings
from tqdm import tqdm

import torch
import torch.nn.functional as F
from torch_geometric.data import Dataset, DataLoader
from torch_geometric.data import Data as PyG_Data


import rdkit
from rdkit import Chem
from rdkit.Chem.rdchem import BondType as BT

from auglichem.utils import (
        ATOM_LIST,
        CHIRALITY_LIST,
        BOND_LIST,
        BONDDIR_LIST,
        random_split,
        scaffold_split
)
from auglichem.molecule import RandomAtomMask, RandomBondDelete, MotifRemoval, Compose
from ._load_sets import read_smiles


#TODO docstrings for MoleculeData

class MoleculeDataset(Dataset):
    def __init__(self, dataset, data_path=None, transform=None, smiles_data=None, labels=None,
                 task=None, test_mode=False, aug_time=0, atom_mask_ratio=[0, 0.25],
                 bond_delete_ratio=[0, 0.25], target=None, class_labels=None, seed=None,
                 augment_original=False, _training_set=False, _train_warn=True, **kwargs):
        '''
            Initialize Molecular Data set object. This object tracks data, labels,
            task, test, and augmentation

            Input:
            -----------------------------------
            dataset (str): Name of data set
            data_path (str, optional default=None): path to store or load data
            smiles_data (np.ndarray of str): Data set in smiles format
            labels (np.ndarray of float): Data labels, maybe optional if unsupervised?
            task (str): 'regression' or 'classification' indicating the learning task
            test_mode (boolean, default=False): Does no augmentations if true
            aug_time (int, optional, default=1): Controls augmentations (not quite sure how yet)
            atom_mask_ratio (list, float): If list, sample mask ratio uniformly over [a, b]
                                           where a < b, if float, set ratio to input.
            bond_delete_ratio (list, float): If list, sample mask ratio uniformly over [a, b]
                                           where a < b, if float, set ratio to input.


            Output:
            -----------------------------------
            None
        '''
        super(Dataset, self).__init__()

        # Store class attributes
        self.dataset = dataset
        self.data_path = data_path

        # Handle incorrectly passed in transformations
        if(isinstance(transform, list)):
            transform = Compose(transform)
        elif(not(isinstance(transform, Compose))):
            transform = Compose([transform])
        self.transform = transform

        if(smiles_data is None):
            self.smiles_data, self.labels, self.task = read_smiles(dataset, data_path)
        else:
            self.smiles_data = smiles_data
            self.labels = labels
            self.task = task

        # No augmentation if no transform is specified
        if(self.transform is None):
            self.test_mode = True
        else:
            self.test_mode = test_mode
            self.aug_time = aug_time

        if self.test_mode:
            self.aug_time = 0

        assert type(self.aug_time) == int

        # For reproducibility
        self.seed = seed
        if(seed is not None):
            np.random.seed(self.seed)
        self.reproduce_seeds = list(range(self.__len__()))
        np.random.shuffle(self.reproduce_seeds)

        # Store mask ratios
        self.atom_mask_ratio = atom_mask_ratio
        self.bond_delete_ratio = bond_delete_ratio

        if(target is not None):
            self.target = target
        else:
            self.target = list(self.labels.keys())[0]

        if(class_labels is not None):
            self.class_labels = class_labels
        else:
            self.class_labels = self.labels[self.target]

        # Augment original data
        self.augment_original = augment_original
        if(self.augment_original):
            warnings.warn("Augmenting original dataset may lead to unexpected results.",
                          RuntimeWarning, stacklevel=2)

        self._training_set = _training_set
        self._train_warn = _train_warn
        self._handle_motifs()


    def _handle_motifs(self):
        # MotifRemoval adds multiple new SMILES strings to our data, and must be done
        # upon training set initialization
        if(self._training_set):
            if(self._train_warn): # Catches if not set through get_data_loaders()
                raise ValueError(
                    "_training_set is for internal use only. " + \
                    "Manually setting _training_set=True is not supported."
                )

            transform = []
            for t in self.transform.transforms:
                if(isinstance(t, MotifRemoval)):
                    print("Gathering Motifs...")
                    new_smiles = []
                    new_labels = []

                    # Match label to each motif
                    for smiles, labels in tqdm(zip(self.smiles_data, self.class_labels)):
                        aug_mols = t(smiles)
                        new_smiles.append(smiles)
                        new_labels.append(labels)
                        for am in aug_mols:
                            new_smiles.append(Chem.MolToSmiles(am))
                            new_labels.append(labels)

                    self.smiles_data = np.array(new_smiles)
                    self.class_labels = np.array(new_labels)
                else:
                    transform.append(t)
            self.transform = Compose(transform)
        else:
            pass


    def _get_data_x(self, mol):
        '''
            Get the transformed data features.

            Inputs:
            -----------------------------------
            mol ( object): Current molecule

            Outputs:
            -----------------------------------
            x (torch.Tensor of longs):
        '''

        # Set up data arrays
        type_idx, chirality_idx, atomic_number = [], [], []

        # Gather atom data
        for atom in mol.GetAtoms():
            try:
                type_idx.append(ATOM_LIST.index(atom.GetAtomicNum()))
                chirality_idx.append(CHIRALITY_LIST.index(atom.GetChiralTag()))
                atomic_number.append(atom.GetAtomicNum())
            except ValueError: # Skip asterisk in motif
                pass

        # Concatenate atom type with chirality index
        x1 = torch.tensor(type_idx, dtype=torch.long).view(-1,1)
        x2 = torch.tensor(chirality_idx, dtype=torch.long).view(-1,1)
        x = torch.cat([x1, x2], dim=-1)

        return x


    def _get_data_y(self, index):
        '''
            Get the transformed data label.

            Inputs:
            -----------------------------------
            index (int): Index for current molecule

            Outputs:
            -----------------------------------
            y (torch.Tensor, long if classification, float if regression): Data label
        '''

        if self.task == 'classification':
            y = torch.tensor(self.class_labels[index], dtype=torch.long).view(1,-1)
        elif self.task == 'regression':
            y = torch.tensor(self.class_labels[index], dtype=torch.float).view(1,-1)

        return y


    def _get_edge_index_and_attr(self, mol):
        '''
            Create the edge index and attributes

            Inputs:
            -----------------------------------
            mol ():

            Outputs:
            -----------------------------------
            edge_index ():
            edge_attr ():
        '''

        # Set up data collection lists
        row, col, edge_feat = [], [], []
        for bond in mol.GetBonds():

            # Get the beginning and end atom indices
            start, end = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()

            # Store bond atoms
            row += [start, end]
            col += [end, start]

            # Store edge featuers
            edge_feat.append([
                BOND_LIST.index(bond.GetBondType()),
                BONDDIR_LIST.index(bond.GetBondDir())
            ])
            edge_feat.append([
                BOND_LIST.index(bond.GetBondType()),
                BONDDIR_LIST.index(bond.GetBondDir())
            ])

        # Create edge index and attributes
        edge_index = torch.tensor([row, col], dtype=torch.long)
        edge_attr = torch.tensor(np.array(edge_feat), dtype=torch.long)

        return edge_index, edge_attr


    def __getitem__(self, index):
        '''
            Selects an element of self.smiles_data according to the index.
            Edge and node masking are done here for each individual molecule

            Input:
            -----------------------------------
            index (int): Index of molecule we would like to augment

            Output:
            -----------------------------------
            masked_data (Data object): data that has been augmented with node and edge masking

        '''
        # If augmentation is done, actual dataset is smaller than given indices
        if self.test_mode:
            true_index = index
        else:
            true_index = index // (self.aug_time + 1)

        # Create initial data set
        mol = Chem.MolFromSmiles(self.smiles_data[true_index])
        mol = Chem.AddHs(mol)

        # Get data x and y
        x = self._get_data_x(mol)
        y = self._get_data_y(true_index)

        # Get edge index and attributes
        edge_index, edge_attr = self._get_edge_index_and_attr(mol)

        # Set up PyG data object
        molecule = PyG_Data(x=x, y=y, edge_index=edge_index, edge_attr=edge_attr,
                            smiles=self.smiles_data[true_index])

        if((not self.test_mode) and (index % (self.aug_time+1) != 0)): # Now retains original
            aug_molecule = self.transform(molecule, seed=self.reproduce_seeds[index])
            return aug_molecule
        elif((not self.test_mode) and self.augment_original):
            aug_molecule = self.transform(molecule, seed=self.reproduce_seeds[index])
            return aug_molecule
        else:
            return molecule


    def _clean_label(self, target):
        """
            Removes labels that don't have values for a given target.
        """
        # If value is not -999999999 we have a valid mol - label pair
        good_idxs = []
        for i, val in enumerate(self.labels[target]):
            if(val != -999999999):
                good_idxs.append(i)
        return good_idxs


    def _match_indices(self, good_idx):
        """
            Match indices between our train/valid/test index and the good indices we have
            for each data subset.
        """

        # For each of train, valid, test, we match where we have a valid molecular
        # representation with where we have a valid label
        updated_train_idx = []
        for v in self.train_idx:
            try:
                updated_train_idx.append(good_idx.index(v))
            except ValueError:
                # No label for training data
                pass

        updated_valid_idx = []
        for v in self.valid_idx:
            try:
                updated_valid_idx.append(good_idx.index(v))
            except ValueError:
                # No label for training data
                pass

        updated_test_idx = []
        for v in self.test_idx:
            try:
                updated_test_idx.append(good_idx.index(v))
            except ValueError:
                # No label for training data
                pass

        # Update indices
        self.train_idx = updated_train_idx
        self.valid_idx = updated_valid_idx
        self.test_idx = updated_test_idx


    def __len__(self):
        return len(self.smiles_data) * (self.aug_time + 1) # Original + augmented


class MoleculeDatasetWrapper(MoleculeDataset):
    def __init__(self, dataset, transform=None, split="scaffold", batch_size=64, num_workers=0,
                 valid_size=0.1, test_size=0.1, aug_time=0, data_path=None, seed=None):
        '''
            Input:
            ---
            dataset (str): One of the datasets available from MoleculeNet
                           (http://moleculenet.ai/datasets-1)
            transform (Compose, OneOf, RandomAtomMask, RandomBondDelete object): transormations
                           to apply to the data at call time.
            split (str, optional default=scaffold): random or scaffold. The splitting strategy
                                                    used for train/test/validation set creation.
            batch_size (int, optional default=64): Batch size used in training
            num_workers (int, optional default=0): Number of workers used in loading data
            valid_size (float in [0,1], optional default=0.1): 
            test_size (float in [0,1],  optional default=0.1): 
            aug_time (int, optional default=0): Number of times to call each augmentation
            data_path (str, optional default=None): specify path to save/lookup data. Default
                        creates `data_download` directory and stores data there
            seed (int, optional, default=None): Random seed to use for reproducibility


            Output:
            ---
            None
        '''
        super().__init__(dataset, data_path, transform, seed=seed)
        self.split = split
        self.data_path = data_path
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.valid_size = valid_size
        self.test_size = test_size
        self.aug_time = aug_time
        self.smiles_data = np.asarray(self.smiles_data)


    def get_data_loaders(self, target=None):
        '''
            Returns torch_geometric DataLoaders for train/validation/test splits. These can
            be iterated over and are ideal for training and evaluating models.

            Inputs:
            -------
            target (str, list of str, optional): Target name to get data loaders for. If None,
                                       returns the loaders for the first target. If 'all'
                                       returns data for all targets at once, ideal for
                                       multitarget trainimg.

            Outputs:
            --------
            train/valid/test_loader (DataLoader): Data loaders containing the train, validation
                                  and test splits of our data.

        '''
        if(not target): # No target passed in, use default and warn
            warnings.warn("No target was set, using {} by default.".format(self.target),
                          RuntimeWarning, stacklevel=2)
            good_idxs = self._clean_label(self.target)
        elif(target == 'all'): # Set to all labels for multitask learning
            self.target = list(self.labels.keys())
            good_idxs = list(range(len(self.smiles_data)))
        elif(isinstance(target, str)): # Set to passed-in single label
            self.target = target
            good_idxs = self._clean_label(self.target)
        elif(isinstance(target, list)): # Multitask on a subset of labels, no cleaning
            self.target = target
            good_idxs = list(range(len(self.smiles_data)))

        # Get indices of data splits
        if(self.split == 'scaffold'):
            self.train_idx, self.valid_idx, self.test_idx = \
                       scaffold_split(self.smiles_data[good_idxs],
                                      self.valid_size, self.test_size)
        elif(self.split == 'random'):
            self.train_idx, self.valid_idx, self.test_idx = \
                       random_split(self.smiles_data[good_idxs],
                                    self.valid_size, self.test_size, self.seed)
        else:
            raise ValueError("Please select scaffold or random split")

        # Need to argwhere for all good_idxs in train, val, test.
        if(isinstance(target, str) and not(target == 'all')):
            self._match_indices(good_idxs)

        # Get data target wither by list or single target
        if(isinstance(self.target, str)):
            labels = np.array(itemgetter(self.target)(self.labels)).T
        else: # Support for multiclass learning
            labels = np.array(itemgetter(*self.target)(self.labels)).T

        # Split into train, validation, and test
        train_set = MoleculeDataset(self.dataset, transform=self.transform,
                            smiles_data=self.smiles_data[good_idxs][self.train_idx],
                            class_labels=labels[good_idxs][self.train_idx],
                            test_mode=False, _training_set=True, _train_warn=False,
                            aug_time=self.aug_time, task=self.task, target=self.target)
        valid_set = MoleculeDataset(self.dataset, transform=self.transform,
                            smiles_data=self.smiles_data[good_idxs][self.valid_idx],
                            class_labels=labels[good_idxs][self.valid_idx],
                            test_mode=True, task=self.task, target=self.target)
        test_set = MoleculeDataset(self.dataset, transform=self.transform,
                           smiles_data=self.smiles_data[good_idxs][self.test_idx],
                           class_labels=labels[good_idxs][self.test_idx],
                           test_mode=True, task=self.task, target=self.target)

        train_loader = DataLoader(
            train_set, batch_size=self.batch_size,
            num_workers=self.num_workers, drop_last=False, shuffle=True
        )

        valid_loader = DataLoader(
            valid_set, batch_size=self.batch_size,
            num_workers=self.num_workers, drop_last=False
        )
        test_loader = DataLoader(
            test_set, batch_size=self.batch_size,
            num_workers=self.num_workers, drop_last=False
        )

        return train_loader, valid_loader, test_loader




from auglichem.crystal.models.cgcnn import CrystalGraphConvNet
from auglichem.crystal.models.schnet import SchNet
from auglichem.crystal.models.gin import GINet
from auglichem.crystal.models.gcn import GCN



# NOTE: Uses code from torchdrug.ai

import pandas as pd
import os
import pathlib
import urllib
from tqdm import tqdm
import csv
import gzip
from rdkit import Chem
import numpy as np
import hashlib

#TODO:
#  1) Clean up code and add proper documentation

USER_AGENT = "auglichem"

def _urlretrieve(url: str, filename: str, chunk_size: int = 1024) -> None:
    with open(filename, "wb") as fh:
        with urllib.request.urlopen(urllib.request.Request(url, headers={"User-Agent": USER_AGENT})) as response:
            with tqdm(total=response.length) as pbar:
                for chunk in iter(lambda: response.read(chunk_size), ""):
                    if not chunk:
                        break
                    pbar.update(chunk_size)
                    fh.write(chunk)

def calculate_md5(fpath: str, chunk_size: int = 1024 * 1024) -> str:
    md5 = hashlib.md5()
    with open(fpath, 'rb') as f:
        for chunk in iter(lambda: f.read(chunk_size), b''):
            md5.update(chunk)
    return md5.hexdigest()


def check_md5(fpath, md5, **kwargs):
    return md5 == calculate_md5(fpath, **kwargs)


def check_integrity(fpath, md5=None):
    if not os.path.isfile(fpath):
        return False
    if md5 is None:
        return True
    return check_md5(fpath, md5)


def _make_csv(fpath, molecules):
    # Open csv from fpath, match SMILES with data, save as new csv
    data_csv = pd.read_csv(fpath)
    smiles = []
    idxs = []
    for idx, m in enumerate(molecules):
        if(m is not None):
            smiles.append(Chem.MolToSmiles(m))
            idxs.append(idx)

    data_csv = data_csv.iloc[idxs]
    data_csv['smiles'] = smiles

    combined_csv_path = fpath[:-8] + ".csv"
    data_csv.to_csv(fpath[:-8] + ".csv")

    return combined_csv_path


def _extract(fpath):
    if(".tar.gz" in fpath):
        import tarfile
        with tarfile.open(fpath, "r") as fin:
            if("gdb8" in fpath):
                fin.extractall(fpath[:-7] + "/")
                fpath = fpath[:-7] + "/qm8.sdf.csv"
                molecules = Chem.SDMolSupplier(fpath[:-4], True, True, False)
            elif("gdb9" in fpath):
                fin.extractall(fpath[:-7] + "/")
                fpath = fpath[:-7] + "/gdb9.sdf.csv"
                molecules = Chem.SDMolSupplier(fpath[:-4], True, True, False)

            # Need to put together smiles and data
            fpath = _make_csv(fpath, molecules)

    elif(".zip" in fpath):
        import zipfile
        with zipfile.ZipFile(fpath) as fin:
            if("FreeSolv" in fpath):
                fin.extractall(fpath[:-4])
                fpath = fpath[:-4] + "/SAMPL.csv"

    return fpath

def download_url(url, root, md5, filename=None):
    """Download a file from a url and place it in root.
    Args:
        url (str): URL to download file from
        root (str): Directory to place downloaded file in
        filename (str, optional): Name to save the file under. If None, use the basename of
                                  the URL
    """
    root = os.path.expanduser(root)
    if not filename:
        filename = os.path.basename(url)
    fpath = os.path.join(root, filename)

    os.makedirs(root, exist_ok=True)

    # Correct file path for extracted sets
    if("FreeSolv.zip" in fpath):
        if(os.path.isfile(fpath[:-4] + "/SAMPL.csv")):
            fpath = fpath[:-4] + "/SAMPL.csv"
    elif("gdb8.tar.gz" in fpath):
        if(os.path.isfile(fpath[:-7] + "/qm8.sdf.csv")):
            fpath = fpath[:-7] + "/qm8.csv"
    elif("gdb9.tar.gz" in fpath):
        if(os.path.isfile(fpath[:-7] + "/gdb9.sdf.csv")):
            fpath = fpath[:-7] + "/gdb9.csv"

    # check if file is already present locally
    if(os.path.isfile(fpath)):
        print("Using: {}".format(fpath))
        return fpath

    # download the file
    try:
        print('Downloading ' + url + ' to ' + fpath)
        _urlretrieve(url, fpath)
    except (urllib.error.URLError, IOError) as e:  # type: ignore[attr-defined]
        if url[:5] == 'https':
            url = url.replace('https:', 'http:')
            print('Failed download. Trying https -> http instead.'
                  ' Downloading ' + url + ' to ' + fpath)
            _urlretrieve(url, fpath)
        else:
            raise e
    
    message = "Download failed or dataset corrupt. Please delete and try again."
    assert check_integrity(fpath, md5), message
    return _extract(fpath)


def _load_data(dataset=None, data_path='./data_download/'):
    if(dataset == 'BACE'):
        task = 'classification'
        target = ['Class']
        md5 = "ba7f8fa3fdf463a811fa7edea8c982c2"
        csv_file_path = download_url(
                    "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/bace.csv",
                    data_path, md5)

    elif(dataset == 'ClinTox'):
        task = 'classification'
        target = ['CT_TOX', 'FDA_APPROVED']
        md5 = "db4f2df08be8ae92814e9d6a2d015284"
        csv_file_path = download_url(
                    "http://deepchem.io.s3-website-us-west-1.amazonaws.com/datasets/clintox.csv.gz",
                    data_path, md5)

    elif(dataset == 'BBBP'):
        task = 'classification'
        target = ['p_np']
        md5 = "66286cb9e6b148bd75d80c870df580fb"
        csv_file_path = download_url(
                   "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/BBBP.csv",
                   data_path, md5)

    elif(dataset == 'SIDER'):
        task = 'classification'
        target = [
            "Hepatobiliary disorders",
            "Metabolism and nutrition disorders",
            "Product issues",
            "Eye disorders",
            "Investigations",
            "Musculoskeletal and connective tissue disorders",
            "Gastrointestinal disorders",
            "Social circumstances",
            "Immune system disorders",
            "Reproductive system and breast disorders",
            "Neoplasms benign, malignant and unspecified (incl cysts and polyps)",
            "General disorders and administration site conditions",
            "Endocrine disorders",
            "Surgical and medical procedures",
            "Vascular disorders",
            "Blood and lymphatic system disorders",
            "Skin and subcutaneous tissue disorders",
            "Congenital, familial and genetic disorders",
            "Infections and infestations",
            "Respiratory, thoracic and mediastinal disorders",
            "Psychiatric disorders",
            "Renal and urinary disorders",
            "Pregnancy, puerperium and perinatal conditions",
            "Ear and labyrinth disorders",
            "Cardiac disorders",
            "Nervous system disorders",
            "Injury, poisoning and procedural complications"
        ]
        md5 = "77c0ef421f7cc8ce963c5836c8761fd2"
        csv_file_path = download_url(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/sider.csv.gz",
                data_path, md5)

    elif(dataset == 'Tox21'):
        task = 'classification'
        target = [
            "NR-AR",
            "NR-AR-LBD",
            "NR-AhR",
            "NR-Aromatase",
            "NR-ER",
            "NR-ER-LBD",
            "NR-PPAR-gamma",
            "SR-ARE",
            "SR-ATAD5",
            "SR-HSE",
            "SR-MMP",
            "SR-p53"
        ]
        md5 = "2882d69e70bba0fec14995f26787cc25"
        csv_file_path = download_url(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/tox21.csv.gz",
                data_path, md5)

    elif(dataset == 'HIV'):
        task = 'classification'
        target = ["HIV_active"]
        md5 = "9ad10c88f82f1dac7eb5c52b668c30a7"
        csv_file_path = download_url(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/HIV.csv",
                data_path, md5)

    elif(dataset == 'MUV'): # Too large to run locally
        task = 'classification'
        target = [
                "MUV-466",
                "MUV-548",
                "MUV-600",
                "MUV-644",
                "MUV-652",
                "MUV-692",
                "MUV-712",
                "MUV-713",
                "MUV-733",
                "MUV-737",
                "MUV-810",
                "MUV-832",
                "MUV-846",
                "MUV-852",
                "MUV-858",
                "MUV-859"
        ]
        md5 = "9c40bd41310991efd40f4d4868fa3ddf"
        csv_file_path = download_url(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/muv.csv.gz",
                data_path, md5)

    elif(dataset == 'FreeSolv'): # SMILES is second column
        task = 'regression'
        target = ["expt"]
        md5 = "8d681babd239b15e2f8b2d29f025577a"
        csv_file_path = download_url(
                "https://s3-us-west-1.amazonaws.com/deepchem.io/datasets/molnet_publish/FreeSolv.zip",
                data_path, md5)

    elif(dataset == 'ESOL'): # SMILES is last column
        task = 'regression'
        target = ["measured log solubility in mols per litre"]
        md5 = "0c90a51668d446b9e3ab77e67662bd1c"
        csv_file_path = download_url(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/delaney-processed.csv",
                data_path, md5)

    elif(dataset == 'Lipophilicity'):
        task = 'regression'
        target = ["exp"]
        md5 = "85a0e1cb8b38b0dfc3f96ff47a57f0ab"
        csv_file_path = download_url(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/Lipophilicity.csv",
                data_path, md5)

    elif(dataset == 'QM7'):
        task = 'regression'
        target = ["u0_atom"]
        md5 = None # No md5 from torchdrug
        csv_file_path = download_url(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm7.csv",
                data_path, md5)

    #elif(dataset == 'QM7b'):
    #    task = 'regression'
    #    target = ["u0_atom"]
    #    csv_file_path = download_url(
    #            "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm7b.mat",
    #            data_path)

    elif(dataset == 'QM8'):
        task = 'regression'
        target = [
                "E1-CC2",
                "E2-CC2",
                "f1-CC2",
                "f2-CC2",
                "E1-PBE0",
                "E2-PBE0",
                "f1-PBE0",
                "f2-PBE0",
                "E1-CAM",
                "E2-CAM",
                "f1-CAM",
                "f2-CAM"
        ]
        md5 = "b7e2a2c823c75b35c596f3013319c86e"
        csv_file_path = download_url(
                "http://deepchem.io.s3-website-us-west-1.amazonaws.com/datasets/gdb8.tar.gz",
                data_path, md5)

    elif(dataset == 'QM9'):
        task = 'regression'
        target = [
                "mu",
                "alpha",
                "homo",
                "lumo",
                "gap",
                "r2",
                "zpve",
                "u0",
                "u298", 
                "h298", 
                "g298", 
                "cv",
        ]
        md5 = "560f62d8e6c992ca0cf8ed8d013f9131"
        csv_file_path = download_url(
                "http://deepchem.io.s3-website-us-west-1.amazonaws.com/datasets/gdb9.tar.gz",
                data_path, md5)

    elif(dataset == "PCBA"):
        task = 'classification'
        target = [
                "PCBA-1030",
                "PCBA-1379",
                "PCBA-1452",
                "PCBA-1454",
                "PCBA-1457",
                "PCBA-1458",
                "PCBA-1460",
                "PCBA-1461",
                "PCBA-1468",
                "PCBA-1469",
                "PCBA-1471",
                "PCBA-1479",
                "PCBA-1631",
                "PCBA-1634",
                "PCBA-1688",
                "PCBA-1721",
                "PCBA-2100",
                "PCBA-2101",
                "PCBA-2147",
                "PCBA-2242",
                "PCBA-2326",
                "PCBA-2451",
                "PCBA-2517",
                "PCBA-2528",
                "PCBA-2546",
                "PCBA-2549",
                "PCBA-2551",
                "PCBA-2662",
                "PCBA-2675",
                "PCBA-2676",
                "PCBA-411",
                "PCBA-463254",
                "PCBA-485281",
                "PCBA-485290",
                "PCBA-485294",
                "PCBA-485297",
                "PCBA-485313",
                "PCBA-485314",
                "PCBA-485341",
                "PCBA-485349",
                "PCBA-485353",
                "PCBA-485360",
                "PCBA-485364",
                "PCBA-485367",
                "PCBA-492947",
                "PCBA-493208",
                "PCBA-504327",
                "PCBA-504332",
                "PCBA-504333",
                "PCBA-504339",
                "PCBA-504444",
                "PCBA-504466",
                "PCBA-504467",
                "PCBA-504706",
                "PCBA-504842",
                "PCBA-504845",
                "PCBA-504847",
                "PCBA-504891",
                "PCBA-540276",
                "PCBA-540317",
                "PCBA-588342",
                "PCBA-588453",
                "PCBA-588456",
                "PCBA-588579",
                "PCBA-588590",
                "PCBA-588591",
                "PCBA-588795",
                "PCBA-588855",
                "PCBA-602179",
                "PCBA-602233",
                "PCBA-602310",
                "PCBA-602313",
                "PCBA-602332",
                "PCBA-624170",
                "PCBA-624171",
                "PCBA-624173",
                "PCBA-624202",
                "PCBA-624246",
                "PCBA-624287",
                "PCBA-624288",
                "PCBA-624291",
                "PCBA-624296",
                "PCBA-624297",
                "PCBA-624417",
                "PCBA-651635",
                "PCBA-651644",
                "PCBA-651768",
                "PCBA-651965",
                "PCBA-652025",
                "PCBA-652104",
                "PCBA-652105",
                "PCBA-652106",
                "PCBA-686970",
                "PCBA-686978",
                "PCBA-686979",
                "PCBA-720504",
                "PCBA-720532",
                "PCBA-720542",
                "PCBA-720551",
                "PCBA-720553",
                "PCBA-720579",
                "PCBA-720580",
                "PCBA-720707",
                "PCBA-720708",
                "PCBA-720709",
                "PCBA-720711",
                "PCBA-743255",
                "PCBA-743266",
                "PCBA-875",
                "PCBA-881",
                "PCBA-883",
                "PCBA-884",
                "PCBA-885",
                "PCBA-887",
                "PCBA-891",
                "PCBA-899",
                "PCBA-902",
                "PCBA-903",
                "PCBA-904",
                "PCBA-912",
                "PCBA-914",
                "PCBA-915",
                "PCBA-924",
                "PCBA-925",
                "PCBA-926",
                "PCBA-927",
                "PCBA-938",
                "PCBA-995",
                ]
        md5 = None # No md5 available from torchdrug
        csv_file_path = download_url(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/pcba.csv.gz",
                data_path, md5)

    #elif(dataset == "PDBbind"):
    #    task = 'regression'
    #    csv_file_path = download_url(
    #"https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/pdbbindv2019/pdbbind_v2019_other_PL.tar.gz",
    #            data_path)
    #    target = ['-logKd/Ki']
    #elif(dataset == "PDBbind - refined"):
    #    task = 'regression'
    #    csv_file_path = download_url(
    #"https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/pdbbindv2019/pdbbind_v2019_refined.tar.gz",
    #            data_path)
    #    target = ['-logKd/Ki']
    #elif(dataset == "PDBbind - core"):
    #    task = 'regression'
    #    csv_file_path = download_url(
    #"https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/pdbbindv2019/pdbbind_v2013_core_set.tar.gz",
    #            data_path)
    #    target = ['-logKd/Ki']

    elif(dataset == 'ToxCast'):
        task = 'classification'
        target = [
                'ACEA_T47D_80hr_Negative', 'ACEA_T47D_80hr_Positive',
                'APR_HepG2_CellCycleArrest_24h_dn', 'APR_HepG2_CellCycleArrest_24h_up',
                'APR_HepG2_CellCycleArrest_72h_dn', 'APR_HepG2_CellLoss_24h_dn',
                'APR_HepG2_CellLoss_72h_dn', 'APR_HepG2_MicrotubuleCSK_24h_dn',
                'APR_HepG2_MicrotubuleCSK_24h_up', 'APR_HepG2_MicrotubuleCSK_72h_dn',
                'APR_HepG2_MicrotubuleCSK_72h_up', 'APR_HepG2_MitoMass_24h_dn',
                'APR_HepG2_MitoMass_24h_up', 'APR_HepG2_MitoMass_72h_dn',
                'APR_HepG2_MitoMass_72h_up', 'APR_HepG2_MitoMembPot_1h_dn',
                'APR_HepG2_MitoMembPot_24h_dn', 'APR_HepG2_MitoMembPot_72h_dn',
                'APR_HepG2_MitoticArrest_24h_up', 'APR_HepG2_MitoticArrest_72h_up',
                'APR_HepG2_NuclearSize_24h_dn', 'APR_HepG2_NuclearSize_72h_dn',
                'APR_HepG2_NuclearSize_72h_up', 'APR_HepG2_OxidativeStress_24h_up',
                'APR_HepG2_OxidativeStress_72h_up', 'APR_HepG2_StressKinase_1h_up',
                'APR_HepG2_StressKinase_24h_up', 'APR_HepG2_StressKinase_72h_up',
                'APR_HepG2_p53Act_24h_up', 'APR_HepG2_p53Act_72h_up',
                'APR_Hepat_Apoptosis_24hr_up', 'APR_Hepat_Apoptosis_48hr_up',
                'APR_Hepat_CellLoss_24hr_dn', 'APR_Hepat_CellLoss_48hr_dn',
                'APR_Hepat_DNADamage_24hr_up', 'APR_Hepat_DNADamage_48hr_up',
                'APR_Hepat_DNATexture_24hr_up', 'APR_Hepat_DNATexture_48hr_up',
                'APR_Hepat_MitoFxnI_1hr_dn', 'APR_Hepat_MitoFxnI_24hr_dn',
                'APR_Hepat_MitoFxnI_48hr_dn', 'APR_Hepat_NuclearSize_24hr_dn',
                'APR_Hepat_NuclearSize_48hr_dn', 'APR_Hepat_Steatosis_24hr_up',
                'APR_Hepat_Steatosis_48hr_up', 'ATG_AP_1_CIS_dn', 'ATG_AP_1_CIS_up',
                'ATG_AP_2_CIS_dn', 'ATG_AP_2_CIS_up', 'ATG_AR_TRANS_dn', 'ATG_AR_TRANS_up',
                'ATG_Ahr_CIS_dn', 'ATG_Ahr_CIS_up', 'ATG_BRE_CIS_dn', 'ATG_BRE_CIS_up',
                'ATG_CAR_TRANS_dn', 'ATG_CAR_TRANS_up', 'ATG_CMV_CIS_dn', 'ATG_CMV_CIS_up',
                'ATG_CRE_CIS_dn', 'ATG_CRE_CIS_up', 'ATG_C_EBP_CIS_dn', 'ATG_C_EBP_CIS_up',
                'ATG_DR4_LXR_CIS_dn', 'ATG_DR4_LXR_CIS_up', 'ATG_DR5_CIS_dn',
                'ATG_DR5_CIS_up', 'ATG_E2F_CIS_dn', 'ATG_E2F_CIS_up', 'ATG_EGR_CIS_up',
                'ATG_ERE_CIS_dn', 'ATG_ERE_CIS_up', 'ATG_ERRa_TRANS_dn',
                'ATG_ERRg_TRANS_dn', 'ATG_ERRg_TRANS_up', 'ATG_ERa_TRANS_up',
                'ATG_E_Box_CIS_dn', 'ATG_E_Box_CIS_up', 'ATG_Ets_CIS_dn', 'ATG_Ets_CIS_up',
                'ATG_FXR_TRANS_up', 'ATG_FoxA2_CIS_dn', 'ATG_FoxA2_CIS_up',
                'ATG_FoxO_CIS_dn', 'ATG_FoxO_CIS_up', 'ATG_GAL4_TRANS_dn',
                'ATG_GATA_CIS_dn', 'ATG_GATA_CIS_up', 'ATG_GLI_CIS_dn', 'ATG_GLI_CIS_up',
                'ATG_GRE_CIS_dn', 'ATG_GRE_CIS_up', 'ATG_GR_TRANS_dn', 'ATG_GR_TRANS_up',
                'ATG_HIF1a_CIS_dn', 'ATG_HIF1a_CIS_up', 'ATG_HNF4a_TRANS_dn',
                'ATG_HNF4a_TRANS_up', 'ATG_HNF6_CIS_dn', 'ATG_HNF6_CIS_up',
                'ATG_HSE_CIS_dn', 'ATG_HSE_CIS_up', 'ATG_IR1_CIS_dn', 'ATG_IR1_CIS_up',
                'ATG_ISRE_CIS_dn', 'ATG_ISRE_CIS_up', 'ATG_LXRa_TRANS_dn',
                'ATG_LXRa_TRANS_up', 'ATG_LXRb_TRANS_dn', 'ATG_LXRb_TRANS_up',
                'ATG_MRE_CIS_up', 'ATG_M_06_TRANS_up', 'ATG_M_19_CIS_dn',
                'ATG_M_19_TRANS_dn', 'ATG_M_19_TRANS_up', 'ATG_M_32_CIS_dn',
                'ATG_M_32_CIS_up', 'ATG_M_32_TRANS_dn', 'ATG_M_32_TRANS_up',
                'ATG_M_61_TRANS_up', 'ATG_Myb_CIS_dn', 'ATG_Myb_CIS_up', 'ATG_Myc_CIS_dn',
                'ATG_Myc_CIS_up', 'ATG_NFI_CIS_dn', 'ATG_NFI_CIS_up', 'ATG_NF_kB_CIS_dn',
                'ATG_NF_kB_CIS_up', 'ATG_NRF1_CIS_dn', 'ATG_NRF1_CIS_up',
                'ATG_NRF2_ARE_CIS_dn', 'ATG_NRF2_ARE_CIS_up', 'ATG_NURR1_TRANS_dn',
                'ATG_NURR1_TRANS_up', 'ATG_Oct_MLP_CIS_dn', 'ATG_Oct_MLP_CIS_up',
                'ATG_PBREM_CIS_dn', 'ATG_PBREM_CIS_up', 'ATG_PPARa_TRANS_dn',
                'ATG_PPARa_TRANS_up', 'ATG_PPARd_TRANS_up', 'ATG_PPARg_TRANS_up',
                'ATG_PPRE_CIS_dn', 'ATG_PPRE_CIS_up', 'ATG_PXRE_CIS_dn', 'ATG_PXRE_CIS_up',
                'ATG_PXR_TRANS_dn', 'ATG_PXR_TRANS_up', 'ATG_Pax6_CIS_up',
                'ATG_RARa_TRANS_dn', 'ATG_RARa_TRANS_up', 'ATG_RARb_TRANS_dn',
                'ATG_RARb_TRANS_up', 'ATG_RARg_TRANS_dn', 'ATG_RARg_TRANS_up',
                'ATG_RORE_CIS_dn', 'ATG_RORE_CIS_up', 'ATG_RORb_TRANS_dn',
                'ATG_RORg_TRANS_dn', 'ATG_RORg_TRANS_up', 'ATG_RXRa_TRANS_dn',
                'ATG_RXRa_TRANS_up', 'ATG_RXRb_TRANS_dn', 'ATG_RXRb_TRANS_up',
                'ATG_SREBP_CIS_dn', 'ATG_SREBP_CIS_up', 'ATG_STAT3_CIS_dn',
                'ATG_STAT3_CIS_up', 'ATG_Sox_CIS_dn', 'ATG_Sox_CIS_up', 'ATG_Sp1_CIS_dn',
                'ATG_Sp1_CIS_up', 'ATG_TAL_CIS_dn', 'ATG_TAL_CIS_up', 'ATG_TA_CIS_dn',
                'ATG_TA_CIS_up', 'ATG_TCF_b_cat_CIS_dn', 'ATG_TCF_b_cat_CIS_up',
                'ATG_TGFb_CIS_dn', 'ATG_TGFb_CIS_up', 'ATG_THRa1_TRANS_dn',
                'ATG_THRa1_TRANS_up', 'ATG_VDRE_CIS_dn', 'ATG_VDRE_CIS_up',
                'ATG_VDR_TRANS_dn', 'ATG_VDR_TRANS_up', 'ATG_XTT_Cytotoxicity_up',
                'ATG_Xbp1_CIS_dn', 'ATG_Xbp1_CIS_up', 'ATG_p53_CIS_dn', 'ATG_p53_CIS_up',
                'BSK_3C_Eselectin_down', 'BSK_3C_HLADR_down', 'BSK_3C_ICAM1_down',
                'BSK_3C_IL8_down', 'BSK_3C_MCP1_down', 'BSK_3C_MIG_down',
                'BSK_3C_Proliferation_down', 'BSK_3C_SRB_down',
                'BSK_3C_Thrombomodulin_down', 'BSK_3C_Thrombomodulin_up',
                'BSK_3C_TissueFactor_down', 'BSK_3C_TissueFactor_up', 'BSK_3C_VCAM1_down',
                'BSK_3C_Vis_down', 'BSK_3C_uPAR_down', 'BSK_4H_Eotaxin3_down',
                'BSK_4H_MCP1_down', 'BSK_4H_Pselectin_down', 'BSK_4H_Pselectin_up',
                'BSK_4H_SRB_down', 'BSK_4H_VCAM1_down', 'BSK_4H_VEGFRII_down',
                'BSK_4H_uPAR_down', 'BSK_4H_uPAR_up', 'BSK_BE3C_HLADR_down',
                'BSK_BE3C_IL1a_down', 'BSK_BE3C_IP10_down', 'BSK_BE3C_MIG_down',
                'BSK_BE3C_MMP1_down', 'BSK_BE3C_MMP1_up', 'BSK_BE3C_PAI1_down',
                'BSK_BE3C_SRB_down', 'BSK_BE3C_TGFb1_down', 'BSK_BE3C_tPA_down',
                'BSK_BE3C_uPAR_down', 'BSK_BE3C_uPAR_up', 'BSK_BE3C_uPA_down',
                'BSK_CASM3C_HLADR_down', 'BSK_CASM3C_IL6_down', 'BSK_CASM3C_IL6_up',
                'BSK_CASM3C_IL8_down', 'BSK_CASM3C_LDLR_down', 'BSK_CASM3C_LDLR_up',
                'BSK_CASM3C_MCP1_down', 'BSK_CASM3C_MCP1_up', 'BSK_CASM3C_MCSF_down',
                'BSK_CASM3C_MCSF_up', 'BSK_CASM3C_MIG_down',
                'BSK_CASM3C_Proliferation_down', 'BSK_CASM3C_Proliferation_up',
                'BSK_CASM3C_SAA_down', 'BSK_CASM3C_SAA_up', 'BSK_CASM3C_SRB_down',
                'BSK_CASM3C_Thrombomodulin_down', 'BSK_CASM3C_Thrombomodulin_up',
                'BSK_CASM3C_TissueFactor_down', 'BSK_CASM3C_VCAM1_down',
                'BSK_CASM3C_VCAM1_up', 'BSK_CASM3C_uPAR_down', 'BSK_CASM3C_uPAR_up',
                'BSK_KF3CT_ICAM1_down', 'BSK_KF3CT_IL1a_down', 'BSK_KF3CT_IP10_down',
                'BSK_KF3CT_IP10_up', 'BSK_KF3CT_MCP1_down', 'BSK_KF3CT_MCP1_up',
                'BSK_KF3CT_MMP9_down', 'BSK_KF3CT_SRB_down', 'BSK_KF3CT_TGFb1_down',
                'BSK_KF3CT_TIMP2_down', 'BSK_KF3CT_uPA_down', 'BSK_LPS_CD40_down',
                'BSK_LPS_Eselectin_down', 'BSK_LPS_Eselectin_up', 'BSK_LPS_IL1a_down',
                'BSK_LPS_IL1a_up', 'BSK_LPS_IL8_down', 'BSK_LPS_IL8_up',
                'BSK_LPS_MCP1_down', 'BSK_LPS_MCSF_down', 'BSK_LPS_PGE2_down',
                'BSK_LPS_PGE2_up', 'BSK_LPS_SRB_down', 'BSK_LPS_TNFa_down',
                'BSK_LPS_TNFa_up', 'BSK_LPS_TissueFactor_down', 'BSK_LPS_TissueFactor_up',
                'BSK_LPS_VCAM1_down', 'BSK_SAg_CD38_down', 'BSK_SAg_CD40_down',
                'BSK_SAg_CD69_down', 'BSK_SAg_Eselectin_down', 'BSK_SAg_Eselectin_up',
                'BSK_SAg_IL8_down', 'BSK_SAg_IL8_up', 'BSK_SAg_MCP1_down',
                'BSK_SAg_MIG_down', 'BSK_SAg_PBMCCytotoxicity_down',
                'BSK_SAg_PBMCCytotoxicity_up', 'BSK_SAg_Proliferation_down',
                'BSK_SAg_SRB_down', 'BSK_hDFCGF_CollagenIII_down', 'BSK_hDFCGF_EGFR_down',
                'BSK_hDFCGF_EGFR_up', 'BSK_hDFCGF_IL8_down', 'BSK_hDFCGF_IP10_down',
                'BSK_hDFCGF_MCSF_down', 'BSK_hDFCGF_MIG_down', 'BSK_hDFCGF_MMP1_down',
                'BSK_hDFCGF_MMP1_up', 'BSK_hDFCGF_PAI1_down',
                'BSK_hDFCGF_Proliferation_down', 'BSK_hDFCGF_SRB_down',
                'BSK_hDFCGF_TIMP1_down', 'BSK_hDFCGF_VCAM1_down', 'CEETOX_H295R_11DCORT_dn',
                'CEETOX_H295R_ANDR_dn', 'CEETOX_H295R_CORTISOL_dn', 'CEETOX_H295R_DOC_dn',
                'CEETOX_H295R_DOC_up', 'CEETOX_H295R_ESTRADIOL_dn',
                'CEETOX_H295R_ESTRADIOL_up', 'CEETOX_H295R_ESTRONE_dn',
                'CEETOX_H295R_ESTRONE_up', 'CEETOX_H295R_OHPREG_up',
                'CEETOX_H295R_OHPROG_dn', 'CEETOX_H295R_OHPROG_up', 'CEETOX_H295R_PROG_up',
                'CEETOX_H295R_TESTO_dn', 'CLD_ABCB1_48hr', 'CLD_ABCG2_48hr',
                'CLD_CYP1A1_24hr', 'CLD_CYP1A1_48hr', 'CLD_CYP1A1_6hr', 'CLD_CYP1A2_24hr',
                'CLD_CYP1A2_48hr', 'CLD_CYP1A2_6hr', 'CLD_CYP2B6_24hr', 'CLD_CYP2B6_48hr',
                'CLD_CYP2B6_6hr', 'CLD_CYP3A4_24hr', 'CLD_CYP3A4_48hr', 'CLD_CYP3A4_6hr',
                'CLD_GSTA2_48hr', 'CLD_SULT2A_24hr', 'CLD_SULT2A_48hr', 'CLD_UGT1A1_24hr',
                'CLD_UGT1A1_48hr', 'NCCT_HEK293T_CellTiterGLO', 'NCCT_QuantiLum_inhib_2_dn',
                'NCCT_QuantiLum_inhib_dn', 'NCCT_TPO_AUR_dn', 'NCCT_TPO_GUA_dn',
                'NHEERL_ZF_144hpf_TERATOSCORE_up', 'NVS_ADME_hCYP19A1', 'NVS_ADME_hCYP1A1',
                'NVS_ADME_hCYP1A2', 'NVS_ADME_hCYP2A6', 'NVS_ADME_hCYP2B6',
                'NVS_ADME_hCYP2C19', 'NVS_ADME_hCYP2C9', 'NVS_ADME_hCYP2D6',
                'NVS_ADME_hCYP3A4', 'NVS_ADME_hCYP4F12', 'NVS_ADME_rCYP2C12',
                'NVS_ENZ_hAChE', 'NVS_ENZ_hAMPKa1', 'NVS_ENZ_hAurA', 'NVS_ENZ_hBACE',
                'NVS_ENZ_hCASP5', 'NVS_ENZ_hCK1D', 'NVS_ENZ_hDUSP3', 'NVS_ENZ_hES',
                'NVS_ENZ_hElastase', 'NVS_ENZ_hFGFR1', 'NVS_ENZ_hGSK3b', 'NVS_ENZ_hMMP1',
                'NVS_ENZ_hMMP13', 'NVS_ENZ_hMMP2', 'NVS_ENZ_hMMP3', 'NVS_ENZ_hMMP7',
                'NVS_ENZ_hMMP9', 'NVS_ENZ_hPDE10', 'NVS_ENZ_hPDE4A1', 'NVS_ENZ_hPDE5',
                'NVS_ENZ_hPI3Ka', 'NVS_ENZ_hPTEN', 'NVS_ENZ_hPTPN11', 'NVS_ENZ_hPTPN12',
                'NVS_ENZ_hPTPN13', 'NVS_ENZ_hPTPN9', 'NVS_ENZ_hPTPRC', 'NVS_ENZ_hSIRT1',
                'NVS_ENZ_hSIRT2', 'NVS_ENZ_hTrkA', 'NVS_ENZ_hVEGFR2', 'NVS_ENZ_oCOX1',
                'NVS_ENZ_oCOX2', 'NVS_ENZ_rAChE', 'NVS_ENZ_rCNOS', 'NVS_ENZ_rMAOAC',
                'NVS_ENZ_rMAOAP', 'NVS_ENZ_rMAOBC', 'NVS_ENZ_rMAOBP', 'NVS_ENZ_rabI2C',
                'NVS_GPCR_bAdoR_NonSelective', 'NVS_GPCR_bDR_NonSelective',
                'NVS_GPCR_g5HT4', 'NVS_GPCR_gH2', 'NVS_GPCR_gLTB4', 'NVS_GPCR_gLTD4',
                'NVS_GPCR_gMPeripheral_NonSelective', 'NVS_GPCR_gOpiateK',
                'NVS_GPCR_h5HT2A', 'NVS_GPCR_h5HT5A', 'NVS_GPCR_h5HT6', 'NVS_GPCR_h5HT7',
                'NVS_GPCR_hAT1', 'NVS_GPCR_hAdoRA1', 'NVS_GPCR_hAdoRA2a',
                'NVS_GPCR_hAdra2A', 'NVS_GPCR_hAdra2C', 'NVS_GPCR_hAdrb1',
                'NVS_GPCR_hAdrb2', 'NVS_GPCR_hAdrb3', 'NVS_GPCR_hDRD1', 'NVS_GPCR_hDRD2s',
                'NVS_GPCR_hDRD4.4', 'NVS_GPCR_hH1', 'NVS_GPCR_hLTB4_BLT1', 'NVS_GPCR_hM1',
                'NVS_GPCR_hM2', 'NVS_GPCR_hM3', 'NVS_GPCR_hM4', 'NVS_GPCR_hNK2',
                'NVS_GPCR_hOpiate_D1', 'NVS_GPCR_hOpiate_mu', 'NVS_GPCR_hTXA2',
                'NVS_GPCR_p5HT2C', 'NVS_GPCR_r5HT1_NonSelective',
                'NVS_GPCR_r5HT_NonSelective', 'NVS_GPCR_rAdra1B',
                'NVS_GPCR_rAdra1_NonSelective', 'NVS_GPCR_rAdra2_NonSelective',
                'NVS_GPCR_rAdrb_NonSelective', 'NVS_GPCR_rNK1', 'NVS_GPCR_rNK3',
                'NVS_GPCR_rOpiate_NonSelective', 'NVS_GPCR_rOpiate_NonSelectiveNa',
                'NVS_GPCR_rSST', 'NVS_GPCR_rTRH', 'NVS_GPCR_rV1', 'NVS_GPCR_rabPAF',
                'NVS_GPCR_rmAdra2B', 'NVS_IC_hKhERGCh', 'NVS_IC_rCaBTZCHL',
                'NVS_IC_rCaDHPRCh_L', 'NVS_IC_rNaCh_site2', 'NVS_LGIC_bGABARa1',
                'NVS_LGIC_h5HT3', 'NVS_LGIC_hNNR_NBungSens', 'NVS_LGIC_rGABAR_NonSelective',
                'NVS_LGIC_rNNR_BungSens', 'NVS_MP_hPBR', 'NVS_MP_rPBR', 'NVS_NR_bER',
                'NVS_NR_bPR', 'NVS_NR_cAR', 'NVS_NR_hAR', 'NVS_NR_hCAR_Antagonist',
                'NVS_NR_hER', 'NVS_NR_hFXR_Agonist', 'NVS_NR_hFXR_Antagonist', 'NVS_NR_hGR',
                'NVS_NR_hPPARa', 'NVS_NR_hPPARg', 'NVS_NR_hPR', 'NVS_NR_hPXR',
                'NVS_NR_hRAR_Antagonist', 'NVS_NR_hRARa_Agonist', 'NVS_NR_hTRa_Antagonist',
                'NVS_NR_mERa', 'NVS_NR_rAR', 'NVS_NR_rMR', 'NVS_OR_gSIGMA_NonSelective',
                'NVS_TR_gDAT', 'NVS_TR_hAdoT', 'NVS_TR_hDAT', 'NVS_TR_hNET', 'NVS_TR_hSERT',
                'NVS_TR_rNET', 'NVS_TR_rSERT', 'NVS_TR_rVMAT2', 'OT_AR_ARELUC_AG_1440',
                'OT_AR_ARSRC1_0480', 'OT_AR_ARSRC1_0960', 'OT_ER_ERaERa_0480',
                'OT_ER_ERaERa_1440', 'OT_ER_ERaERb_0480', 'OT_ER_ERaERb_1440',
                'OT_ER_ERbERb_0480', 'OT_ER_ERbERb_1440', 'OT_ERa_EREGFP_0120',
                'OT_ERa_EREGFP_0480', 'OT_FXR_FXRSRC1_0480', 'OT_FXR_FXRSRC1_1440',
                'OT_NURR1_NURR1RXRa_0480', 'OT_NURR1_NURR1RXRa_1440',
                'TOX21_ARE_BLA_Agonist_ch1', 'TOX21_ARE_BLA_Agonist_ch2',
                'TOX21_ARE_BLA_agonist_ratio', 'TOX21_ARE_BLA_agonist_viability',
                'TOX21_AR_BLA_Agonist_ch1', 'TOX21_AR_BLA_Agonist_ch2',
                'TOX21_AR_BLA_Agonist_ratio', 'TOX21_AR_BLA_Antagonist_ch1',
                'TOX21_AR_BLA_Antagonist_ch2', 'TOX21_AR_BLA_Antagonist_ratio',
                'TOX21_AR_BLA_Antagonist_viability', 'TOX21_AR_LUC_MDAKB2_Agonist',
                'TOX21_AR_LUC_MDAKB2_Antagonist', 'TOX21_AR_LUC_MDAKB2_Antagonist2',
                'TOX21_AhR_LUC_Agonist', 'TOX21_Aromatase_Inhibition',
                'TOX21_AutoFluor_HEK293_Cell_blue', 'TOX21_AutoFluor_HEK293_Media_blue',
                'TOX21_AutoFluor_HEPG2_Cell_blue', 'TOX21_AutoFluor_HEPG2_Cell_green',
                'TOX21_AutoFluor_HEPG2_Media_blue', 'TOX21_AutoFluor_HEPG2_Media_green',
                'TOX21_ELG1_LUC_Agonist', 'TOX21_ERa_BLA_Agonist_ch1',
                'TOX21_ERa_BLA_Agonist_ch2', 'TOX21_ERa_BLA_Agonist_ratio',
                'TOX21_ERa_BLA_Antagonist_ch1', 'TOX21_ERa_BLA_Antagonist_ch2',
                'TOX21_ERa_BLA_Antagonist_ratio', 'TOX21_ERa_BLA_Antagonist_viability',
                'TOX21_ERa_LUC_BG1_Agonist', 'TOX21_ERa_LUC_BG1_Antagonist',
                'TOX21_ESRE_BLA_ch1', 'TOX21_ESRE_BLA_ch2', 'TOX21_ESRE_BLA_ratio',
                'TOX21_ESRE_BLA_viability', 'TOX21_FXR_BLA_Antagonist_ch1',
                'TOX21_FXR_BLA_Antagonist_ch2', 'TOX21_FXR_BLA_agonist_ch2',
                'TOX21_FXR_BLA_agonist_ratio', 'TOX21_FXR_BLA_antagonist_ratio',
                'TOX21_FXR_BLA_antagonist_viability', 'TOX21_GR_BLA_Agonist_ch1',
                'TOX21_GR_BLA_Agonist_ch2', 'TOX21_GR_BLA_Agonist_ratio',
                'TOX21_GR_BLA_Antagonist_ch2', 'TOX21_GR_BLA_Antagonist_ratio',
                'TOX21_GR_BLA_Antagonist_viability', 'TOX21_HSE_BLA_agonist_ch1',
                'TOX21_HSE_BLA_agonist_ch2', 'TOX21_HSE_BLA_agonist_ratio',
                'TOX21_HSE_BLA_agonist_viability', 'TOX21_MMP_ratio_down',
                'TOX21_MMP_ratio_up', 'TOX21_MMP_viability', 'TOX21_NFkB_BLA_agonist_ch1',
                'TOX21_NFkB_BLA_agonist_ch2', 'TOX21_NFkB_BLA_agonist_ratio',
                'TOX21_NFkB_BLA_agonist_viability', 'TOX21_PPARd_BLA_Agonist_viability',
                'TOX21_PPARd_BLA_Antagonist_ch1', 'TOX21_PPARd_BLA_agonist_ch1',
                'TOX21_PPARd_BLA_agonist_ch2', 'TOX21_PPARd_BLA_agonist_ratio',
                'TOX21_PPARd_BLA_antagonist_ratio', 'TOX21_PPARd_BLA_antagonist_viability',
                'TOX21_PPARg_BLA_Agonist_ch1', 'TOX21_PPARg_BLA_Agonist_ch2',
                'TOX21_PPARg_BLA_Agonist_ratio', 'TOX21_PPARg_BLA_Antagonist_ch1',
                'TOX21_PPARg_BLA_antagonist_ratio', 'TOX21_PPARg_BLA_antagonist_viability',
                'TOX21_TR_LUC_GH3_Agonist', 'TOX21_TR_LUC_GH3_Antagonist',
                'TOX21_VDR_BLA_Agonist_viability', 'TOX21_VDR_BLA_Antagonist_ch1',
                'TOX21_VDR_BLA_agonist_ch2', 'TOX21_VDR_BLA_agonist_ratio',
                'TOX21_VDR_BLA_antagonist_ratio', 'TOX21_VDR_BLA_antagonist_viability',
                'TOX21_p53_BLA_p1_ch1', 'TOX21_p53_BLA_p1_ch2', 'TOX21_p53_BLA_p1_ratio',
                'TOX21_p53_BLA_p1_viability', 'TOX21_p53_BLA_p2_ch1', 'TOX21_p53_BLA_p2_ch2',
                'TOX21_p53_BLA_p2_ratio', 'TOX21_p53_BLA_p2_viability',
                'TOX21_p53_BLA_p3_ch1', 'TOX21_p53_BLA_p3_ch2', 'TOX21_p53_BLA_p3_ratio',
                'TOX21_p53_BLA_p3_viability', 'TOX21_p53_BLA_p4_ch1', 'TOX21_p53_BLA_p4_ch2',
                'TOX21_p53_BLA_p4_ratio', 'TOX21_p53_BLA_p4_viability',
                'TOX21_p53_BLA_p5_ch1', 'TOX21_p53_BLA_p5_ch2', 'TOX21_p53_BLA_p5_ratio',
                'TOX21_p53_BLA_p5_viability', 'Tanguay_ZF_120hpf_AXIS_up',
                'Tanguay_ZF_120hpf_ActivityScore', 'Tanguay_ZF_120hpf_BRAI_up',
                'Tanguay_ZF_120hpf_CFIN_up', 'Tanguay_ZF_120hpf_CIRC_up',
                'Tanguay_ZF_120hpf_EYE_up', 'Tanguay_ZF_120hpf_JAW_up',
                'Tanguay_ZF_120hpf_MORT_up', 'Tanguay_ZF_120hpf_OTIC_up',
                'Tanguay_ZF_120hpf_PE_up', 'Tanguay_ZF_120hpf_PFIN_up',
                'Tanguay_ZF_120hpf_PIG_up', 'Tanguay_ZF_120hpf_SNOU_up',
                'Tanguay_ZF_120hpf_SOMI_up', 'Tanguay_ZF_120hpf_SWIM_up',
                'Tanguay_ZF_120hpf_TRUN_up', 'Tanguay_ZF_120hpf_TR_up',
                'Tanguay_ZF_120hpf_YSE_up'
            ]
        md5 = "92911bbf9c1e2ad85231014859388cd6"
        csv_file_path = download_url(
                "http://deepchem.io.s3-website-us-west-1.amazonaws.com/datasets/toxcast_data.csv.gz",
                data_path, md5)

    else:
        raise RuntimeError("{} is not supported.".format(dataset))
    print("DATASET: {}".format(dataset))
    return csv_file_path, target, task


def _process_csv(csv_file, target, task):
    csv_reader = csv.DictReader(csv_file, delimiter=',')
    smiles_data, labels = [], {}
    for t in target:
        labels[t] = []

    for i, row in tqdm(enumerate(csv_reader)):
        # Skip header
        if i == 0:
            continue

        # Hold on to all rows that contain any targets
        try:
            smiles = row['smiles']
        except KeyError:
            smiles = row['mol']

        # Hold on to good smiles representations
        mol = Chem.MolFromSmiles(smiles) 
        if(mol != None):
            smiles_data.append(smiles)
        else:
            continue

        # Check values for every target
        for idx, t in enumerate(target):
            label = row[t]
            if label != '':
                if task == 'classification':
                    try:
                        labels[t].append(int(label))
                    except ValueError:
                        labels[t].append(int(label[0]))
                elif task == 'regression':
                    labels[t].append(float(label))
                else:
                    raise ValueError('task must be either regression or classification')
            elif mol != None and label == '':
                labels[t].append(-999999999)

    # Recast lables to numpy arrays
    for t in target:
        labels[t] = np.array(labels[t])

    #for key, val in labels.items():
    #    print(key, len(val))

    return smiles_data, labels, task


def read_smiles(dataset, data_path):

    # Create data download directory if it does not exist
    if(data_path is None):
        data_path = './data_download'
    if(not os.path.exists(data_path)):
        os.mkdir(data_path)

    # Download files if not already there
    csv_file_path, target, task = _load_data(dataset, data_path)
    if("csv.gz" in csv_file_path):
        with gzip.open(csv_file_path, 'rt') as csv_file:
            return _process_csv(csv_file, target, task)
    else:
        with open(csv_file_path) as csv_file:
            return _process_csv(csv_file, target, task)

    return smiles_data, np.array(labels)




from rdkit import Chem
from rdkit.Chem.Scaffolds.MurckoScaffold import MurckoScaffoldSmiles
import random

def random_split(dataset, valid_size, test_size, seed=None):

    # Set seed
    if(seed is not None):
        random.seed(seed)

    # Get indices
    total_size = len(dataset)
    indices = list(range(total_size))
    random.shuffle(indices)

    train_size = 1.0 - valid_size - test_size

    train_cutoff = int(train_size * total_size)
    valid_cutoff = int((train_size + valid_size) * total_size)

    train_idx = indices[:train_cutoff]
    valid_idx = indices[train_cutoff:valid_cutoff]
    test_idx = indices[valid_cutoff:]
    return train_idx, valid_idx, test_idx


def _generate_scaffold(smiles, include_chirality=False):
    mol = Chem.MolFromSmiles(smiles)
    scaffold = MurckoScaffoldSmiles(mol=mol, includeChirality=include_chirality)
    return scaffold


def generate_scaffolds(smiles_data, log_every_n=1000):
    scaffolds = {}
    data_len = len(smiles_data)

    print("Generating scaffolds...")
    for ind, smiles in enumerate(smiles_data):
        if ind % log_every_n == 0:
            print("Generating scaffold %d/%d" % (ind, data_len))
        scaffold = _generate_scaffold(smiles)
        if scaffold not in scaffolds:
            scaffolds[scaffold] = [ind]
        else:
            scaffolds[scaffold].append(ind)

    # Sort from largest to smallest scaffold sets
    scaffolds = {key: sorted(value) for key, value in scaffolds.items()}
    scaffold_sets = [
        scaffold_set for (scaffold, scaffold_set) in sorted(
            scaffolds.items(), key=lambda x: (len(x[1]), x[1][0]), reverse=True)
    ]
    return scaffold_sets


def scaffold_split(smiles_data, valid_size, test_size, log_every_n=1000):
    train_size = 1.0 - valid_size - test_size
    scaffold_sets = generate_scaffolds(smiles_data)

    train_cutoff = train_size * len(smiles_data)
    valid_cutoff = (train_size + valid_size) * len(smiles_data)
    train_inds: List[int] = []
    valid_inds: List[int] = []
    test_inds: List[int] = []

    print("About to sort in scaffold sets")
    for scaffold_set in scaffold_sets:
        if len(train_inds) + len(scaffold_set) > train_cutoff:
            if len(train_inds) + len(valid_inds) + len(scaffold_set) > valid_cutoff:
                test_inds += scaffold_set
            else:
                valid_inds += scaffold_set
        else:
            train_inds += scaffold_set
    return train_inds, valid_inds, test_inds



from auglichem.molecule.models.gcn import GCN
from auglichem.molecule.models.deepgcn import DeepGCN
from auglichem.molecule.models.gine import GINE
from auglichem.molecule.models.afp import AttentiveFP



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.

import random
from typing import Any, Dict, List, Optional

from auglichem.molecule._transforms import BaseTransform

"""
Composition Operators:
Compose: the Compose operator was added here such that users
do not have to import `torchvision` in order to compose multiple
augmentations together. These operators work identically and either
can be used.
OneOf: the OneOf operator takes as input a list of transforms and
may apply (with probability p) one of the transforms in the list.
If a transform is applied, it is selected using the specified
probabilities of the individual transforms.
Example:
 >>> Compose([
 >>>     Blur(),
 >>>     ColorJitter(saturation_factor=1.5),
 >>>     OneOf([
 >>>         OverlayOntoScreenshot(),
 >>>         OverlayEmoji(),
 >>>         OverlayText(),
 >>>     ]),
 >>> ])
"""


class BaseComposition(object):
    def __init__(self, transforms: List[BaseTransform], p: float = 1.0):
        """
        @param transforms: a list of transforms
        @param p: the probability of the transform being applied; default value is 1.0
        """
        for transform in transforms:
            assert isinstance(
                transform, (BaseTransform, BaseComposition, object)
            ), "Expected instances of type `BaseTransform` or `BaseComposition` for variable `transforms`"  # noqa: B950
        assert 0 <= p <= 1.0, "p must be a value in the range [0, 1]"

        self.transforms = transforms
        self.p = p


class Compose(BaseComposition):
    def __call__(self, molecule, seed=None):
        """
        Applies the list of transforms in order to the molecule
        @param molecule: PIL Image to be augmented
        @param metadata: if set to be a list, metadata about the function execution
            including its name, the source & dest width, height, etc. will be appended to
            the inputted list. If set to None, no metadata will be appended or returned
        @returns: Augmented PIL Image
        """
        if random.random() > self.p:
            return molecule

        for transform in self.transforms:
            molecule = transform(molecule, seed=seed)

        return molecule
    
    def __str__(self):
        print_str = "Compose object with transformations:"
        for t in self.transforms:
            print_str += "\n\t" + str(t)
        return print_str


class OneOf(BaseComposition):
    def __init__(self, transforms: List[BaseTransform], p: float = 1.0):
        """
        @param transforms: a list of transforms to select from; one of which
            will be chosen to be applied to the media
        @param p: the probability of the transform being applied; default value is 1.0
        """
        super().__init__(transforms, p)
        transform_probs = [t.p for t in transforms]
        probs_sum = sum(transform_probs)
        self.transform_probs = [t / probs_sum for t in transform_probs]

    def __call__(self, molecule, seed=None):
        """
        Applies one of the transforms to the molecule
        @param molecule: PIL Image to be augmented
        @param metadata: if set to be a list, metadata about the function execution
            including its name, the source & dest width, height, etc. will be appended to
            the inputted list. If set to None, no metadata will be appended or returned
        @returns: Augmented PIL Image
        """
        if random.random() > self.p:
            return molecule

        transform = random.choices(self.transforms, self.transform_probs)[0]
        return transform(molecule, force=True, metadata=metadata)



from typing import Optional
from torch_geometric.typing import Adj, OptTensor

import torch
from torch import Tensor, nn
import torch.nn.functional as F
from torch.nn import Linear, Parameter, GRUCell

from torch_geometric.utils import softmax, add_self_loops
from torch_geometric.nn import GATConv, MessagePassing, global_add_pool
from torch_geometric.nn.inits import glorot, zeros

#from data_aug.dataset import ATOM_LIST, CHIRALITY_LIST, BOND_LIST, BONDDIR_LIST
from auglichem.utils import NUM_ATOM_TYPE, NUM_CHIRALITY_TAG, NUM_BOND_TYPE, NUM_BOND_DIRECTION

#NUM_ATOM_TYPE = len(ATOM_LIST) + 1 # including the extra mask tokens
#NUM_CHIRALITY_TAG = len(CHIRALITY_LIST)
#NUM_BOND_TYPE = len(BOND_LIST) + 1 # including aromatic and self-loop edge
#NUM_BOND_DIRECTION = len(BONDDIR_LIST)


class GATEConv(MessagePassing):
    def __init__(self, emb_dim, dropout=0.0):
        super(GATEConv, self).__init__(aggr='add', node_dim=0)

        self.dropout = dropout

        self.edge_embedding1 = nn.Embedding(NUM_BOND_TYPE, emb_dim)
        self.edge_embedding2 = nn.Embedding(NUM_BOND_DIRECTION, emb_dim)
        nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        nn.init.xavier_uniform_(self.edge_embedding2.weight.data)

        self.att_l = Parameter(torch.Tensor(1, emb_dim))
        self.att_r = Parameter(torch.Tensor(1, emb_dim))

        self.lin1 = Linear(emb_dim*2, emb_dim, False)
        self.lin2 = Linear(emb_dim, emb_dim, False)

        self.bias = Parameter(torch.Tensor(emb_dim))

        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.att_l)
        glorot(self.att_r)
        glorot(self.lin1.weight)
        glorot(self.lin2.weight)
        zeros(self.bias)

    def forward(self, x, edge_index, edge_attr):
        edge_index = add_self_loops(edge_index, num_nodes=x.size(0))[0]

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:,0] = 4 #bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        edge_embeddings = self.edge_embedding1(edge_attr[:,0]) + self.edge_embedding2(edge_attr[:,1])

        out = self.propagate(edge_index, x=x, edge_attr=edge_embeddings)
        out += self.bias
        return out

    def message(self, x_j, x_i, edge_attr,
                index, ptr, size_i: Optional[int]):
        x_j = F.leaky_relu_(self.lin1(torch.cat([x_j, edge_attr], dim=-1)))
        alpha_j = (x_j * self.att_l).sum(dim=-1)
        alpha_i = (x_i * self.att_r).sum(dim=-1)
        alpha = alpha_j + alpha_i
        alpha = F.leaky_relu_(alpha)
        alpha = softmax(alpha, index, ptr, size_i)
        alpha = F.dropout(alpha, p=self.dropout, training=self.training)
        return self.lin2(x_j) * alpha.unsqueeze(-1)


class AttentiveFP(nn.Module):
    """
    Args:
        edge_dim (int): Edge feature dimensionality.
        num_layers (int): Number of GNN layers.
        num_timesteps (int): Number of iterative refinement steps for global
            readout.
        dropout (float, optional): Dropout probability. (default: :obj:`0.0`)
    """
    def __init__(self, task, emb_dim=300, num_layers=5, num_timesteps=3, drop_ratio=0,
                 output_dim=None, **kwargs):
        super(AttentiveFP, self).__init__()
        if(output_dim == None):
            self.output_dim = 1
        else:
            self.output_dim = output_dim


        self.x_embedding1 = nn.Embedding(NUM_ATOM_TYPE, emb_dim)
        self.x_embedding2 = nn.Embedding(NUM_CHIRALITY_TAG, emb_dim)
        nn.init.xavier_uniform_(self.x_embedding1.weight.data)
        nn.init.xavier_uniform_(self.x_embedding2.weight.data)

        self.task = task
        self.num_layers = num_layers
        self.num_timesteps = num_timesteps
        self.drop_ratio = drop_ratio

        self.lin1 = Linear(emb_dim, emb_dim)

        conv = GATEConv(emb_dim, drop_ratio)
        gru = GRUCell(emb_dim, emb_dim)
        self.atom_convs = torch.nn.ModuleList([conv])
        self.atom_grus = torch.nn.ModuleList([gru])
        for _ in range(num_layers - 1):
            conv = GATConv(emb_dim, emb_dim, dropout=drop_ratio,
                           add_self_loops=False, )#negative_slope=0.01)
            self.atom_convs.append(conv)
            self.atom_grus.append(GRUCell(emb_dim, emb_dim))

        self.mol_conv = GATConv(emb_dim, emb_dim,
                                dropout=drop_ratio, add_self_loops=False,
                                )#negative_slope=0.01)
        self.mol_gru = GRUCell(emb_dim, emb_dim)

        # self.lin2 = Linear(emb_dim, out_channels)
        if self.task == 'classification':
            self.output_dim *= 2
            self.lin2 = nn.Linear(emb_dim, self.output_dim)
        elif self.task == 'regression':
            self.lin2 = nn.Linear(emb_dim, self.output_dim)

        self.reset_parameters()

    def reset_parameters(self):
        self.lin1.reset_parameters()
        for conv, gru in zip(self.atom_convs, self.atom_grus):
            conv.reset_parameters()
            gru.reset_parameters()
        self.mol_conv.reset_parameters()
        self.mol_gru.reset_parameters()
        self.lin2.reset_parameters()

    def forward(self, data):
        """"""
        x = data.x
        edge_index = data.edge_index
        edge_attr = data.edge_attr

        x = self.x_embedding1(x[:,0]) + self.x_embedding2(x[:,1])

        # Atom Embedding:
        x = F.leaky_relu_(self.lin1(x))

        h = F.elu_(self.atom_convs[0](x, edge_index, edge_attr))
        h = F.dropout(h, p=self.drop_ratio, training=self.training)
        x = self.atom_grus[0](h, x).relu_()

        for conv, gru in zip(self.atom_convs[1:], self.atom_grus[1:]):
            h = F.elu_(conv(x, edge_index))
            h = F.dropout(h, p=self.drop_ratio, training=self.training)
            x = gru(h, x).relu_()

        # Molecule Embedding:
        row = torch.arange(data.batch.size(0), device=data.batch.device)
        edge_index = torch.stack([row, data.batch], dim=0)

        out = global_add_pool(x, data.batch).relu_()
        for t in range(self.num_timesteps):
            h = F.elu_(self.mol_conv((x, out), edge_index))
            h = F.dropout(h, p=self.drop_ratio, training=self.training)
            out = self.mol_gru(h, out).relu_()

        # Predictor:
        out = F.dropout(out, p=self.drop_ratio, training=self.training)
        return out, self.lin2(out)



from __future__ import print_function, division

import sys
import csv
import functools
import json
import os
import random
import ase
from ase.io import read, write
import warnings
import numpy as np
from pymatgen.core.structure import Structure
import glob
from ase.io import  read, write
from ase import Atoms,Atom
from pymatgen.core.operations import SymmOp
from pymatgen.transformations.transformation_abc import AbstractTransformation
import math
from typing import Optional, Union
from pymatgen.core.structure import Molecule, Structure


class AseAtomsAdaptor:
    """
    Adaptor serves as a bridge between ASE Atoms and pymatgen objects.
    """

    @staticmethod
    def get_atoms(structure, **kwargs):
        """
        Returns ASE Atoms object from pymatgen structure or molecule.
        Args:
            structure: pymatgen.core.structure.Structure or pymatgen.core.structure.Molecule
            **kwargs: other keyword args to pass into the ASE Atoms constructor
        Returns:
            ASE Atoms object
        """
        if not structure.is_ordered:
            raise ValueError("ASE Atoms only supports ordered structures")
        #if not ase_loaded:
        if('ase' not in sys.modules):
            raise ImportError(
                "AseAtomsAdaptor requires ase package.\n" "Use `pip install ase` or `conda install ase -c conda-forge`"
            )
        symbols = [str(site.specie.symbol) for site in structure]
        positions = [site.coords for site in structure]
        if hasattr(structure, "lattice"):
            cell = structure.lattice.matrix
            pbc = True
        else:
            cell = None
            pbc = None
        return Atoms(symbols=symbols, positions=positions, pbc=pbc, cell=cell, **kwargs)

    @staticmethod
    def get_structure(atoms, cls=None):
        """
        Returns pymatgen structure from ASE Atoms.
        Args:
            atoms: ASE Atoms object
            cls: The Structure class to instantiate (defaults to pymatgen structure)
        Returns:
            Equivalent pymatgen.core.structure.Structure
        """
        symbols = atoms.get_chemical_symbols()
        positions = atoms.get_positions()
        lattice = atoms.get_cell()

        cls = Structure if cls is None else cls
        return cls(lattice, symbols, positions, coords_are_cartesian=True)


class RotationTransformation(AbstractTransformation):
    """
    The RotationTransformation applies a rotation to a structure.
    """

    def __init__(self, axis=None, angle=None, perturb=None, angle_in_radians=False):
        """
        Args:
            axis (3x1 array): Axis of rotation, e.g., [1, 0, 0]
            angle (float): Angle to rotate
            perturb (PerturbStructureTransformation): pre-rotation perturbation
            angle_in_radians (boolean): Flag for angle in radians instead of degrees
        """
        self.axis = axis
        self.angle = angle
        self.perturb = perturb
        self.angle_in_radians=False

    def apply_transformation(self, structure, axis=None, angle=None, perturb=None,
                             angle_in_radians=False, seed=None):
        """
        Apply the transformation.
        Args:
            structure (Structure): Input Structure
        Returns:
            Rotated Structure.
        """
        # Set random seed
        if(seed is not None):
            np.random.seed(seed)

        # Read in new parameters
        if(axis is not None):
            self.axis = axis
        if(angle is not None):
            self.angle = angle

        # Apply perturbation if passed in
        if(perturb is not None):
            self.perturb = perturb
        if(self.perturb is not None):
            structure = self.perturb.apply_transformation(structure)

        # Randomly generate an angle
        if(self.angle is None):
            self.angle = np.random.choice(360, 1, replace=False)

        self.angle_in_radians = angle_in_radians
        self._symmop = SymmOp.from_axis_angle_and_translation(self.axis, self.angle, self.angle_in_radians)

        s = structure.copy()
        s.apply_operation(self._symmop, fractional=True)
        # s = self._symmop(s)
        return s

    def __str__(self):
        return "Rotation Transformation about axis " + "{} with angle = {:.4f} {}".format(
            self.axis, self.angle, "radians" if self.angle_in_radians else "degrees"
        )

    def __repr__(self):
        return self.__str__()

    @property
    def inverse(self):
        """
        Returns:
            Inverse Transformation.
        """
        return RotationTransformation(self.axis, -self.angle, self.angle_in_radians)

    @property
    def is_one_to_many(self):
        """Returns: False"""
        return False


class PerturbStructureTransformation(AbstractTransformation):
    """
    This transformation perturbs a structure by a specified distance in random
    directions. Used for breaking symmetries.
    """

    def __init__(
        self,
        distance: float = 0.01,
        min_distance: Optional[Union[int, float]] = None,
    ):
        """
        Args:
            distance: Distance of perturbation in angstroms. All sites
                will be perturbed by exactly that distance in a random
                direction.
            min_distance: if None, all displacements will be equidistant. If int
                or float, perturb each site a distance drawn from the uniform
                distribution between 'min_distance' and 'distance'.
        """
        self.distance = distance
        self.min_distance = min_distance

    def apply_transformation(self, structure: Structure, seed=None) -> Structure:
        """
        Apply the transformation.
        Args:
            structure: Input Structure
        Returns:
            Structure with sites perturbed.
        """
        s = structure.copy()
        if(seed is not None):
            np.random.seed(seed) # Pymatgen uses numpy random
        s.perturb(self.distance, min_distance=self.min_distance)
        return s

    def __str__(self):
        return "PerturbStructureTransformation : " + "Min_distance = {}".format(self.min_distance)

    def __repr__(self):
        return self.__str__()

    @property
    def inverse(self):
        """
        Returns: None
        """
        return None

    @property
    def is_one_to_many(self):
        """
        Returns: False
        """
        return False


class SwapAxesTransformation(object):
    """
    Swap coordinate axes in the crystal strucutre.
    """
    def __init__(self):
        #TODO: Shouldn't this always call the transformation?
        #TODO: Decide if its probabilistic or not, and if we should use random seeding.
        pass

    def apply_transformation(self, crys, seed=None, _test_choice=None):
        atoms = AseAtomsAdaptor().get_atoms(crys)

        if(seed is not None):
            np.random.seed(seed)
        choice = np.random.choice(3, 2, replace=False)
        if(_test_choice is not None):
            choice = _test_choice
        pos = (atoms.positions)
        pos[:,[choice[0], choice[1]]] = pos[:,[choice[1], choice[0]]]
        atoms.arrays["positions"] = pos

        return AseAtomsAdaptor.get_structure(atoms)

    def __str__(self):
        return "SwapAxesTransformation"

    def __repr__(self):
        return self.__str__()

class RemoveSitesTransformation(AbstractTransformation):
    """
    Remove certain sites in a structure.
    """

    def __init__(self, indices_to_remove):
        """
        Args:
            indices_to_remove: List of indices to remove. E.g., [0, 1, 2]
        """

        self.indices_to_remove = indices_to_remove

    def apply_transformation(self, structure, seed=None):
        """
        Apply the transformation.
        Arg:
            structure (Structure): A structurally similar structure in
                regards to crystal and site positions.
        Return:
            Returns a copy of structure with sites removed.
        """
        s = structure.copy()
        s.remove_sites(self.indices_to_remove)
        return s

    def __str__(self):
        return "RemoveSitesTransformation :" + ", ".join(map(str, self.indices_to_remove))

    def __repr__(self):
        return self.__str__()

    @property
    def inverse(self):
        """Return: None"""
        return None

    @property
    def is_one_to_many(self):
        """Return: False"""
        return False


def _proj(b, a):
    """
    Returns vector projection (np.ndarray) of vector b (np.ndarray)
    onto vector a (np.ndarray)
    """
    return (b.T @ (a / np.linalg.norm(a))) * (a / np.linalg.norm(a))

def _round_and_make_arr_singular(arr: np.ndarray) -> np.ndarray:
    """
    This function rounds all elements of a matrix to the nearest integer,
    unless the rounding scheme causes the matrix to be singular, in which
    case elements of zero rows or columns in the rounded matrix with the
    largest absolute valued magnitude in the unrounded matrix will be
    rounded to the next integer away from zero rather than to the
    nearest integer.
    The transformation is as follows. First, all entries in 'arr' will be
    rounded to the nearest integer to yield 'arr_rounded'. If 'arr_rounded'
    has any zero rows, then one element in each zero row of 'arr_rounded'
    corresponding to the element in 'arr' of that row with the largest
    absolute valued magnitude will be rounded to the next integer away from
    zero (see the '_round_away_from_zero(x)' function) rather than the
    nearest integer. This process is then repeated for zero columns. Also
    note that if 'arr' already has zero rows or columns, then this function
    will not change those rows/columns.
    Args:
        arr: Input matrix
    Returns:
        Transformed matrix.
    """

    def round_away_from_zero(x):
        """
        Returns 'x' rounded to the next integer away from 0.
        If 'x' is zero, then returns zero.
        E.g. -1.2 rounds to -2.0. 1.2 rounds to 2.0.
        """
        abs_x = abs(x)
        return math.ceil(abs_x) * (abs_x / x) if x != 0 else 0

    arr_rounded = np.around(arr)

    # Zero rows in 'arr_rounded' make the array singular, so force zero rows to
    # be nonzero
    if (~arr_rounded.any(axis=1)).any():
        # Check for zero rows in T_rounded

        # indices of zero rows
        zero_row_idxs = np.where(~arr_rounded.any(axis=1))[0]

        for zero_row_idx in zero_row_idxs:  # loop over zero rows
            zero_row = arr[zero_row_idx, :]

            # Find the element of the zero row with the largest absolute
            # magnitude in the original (non-rounded) array (i.e. 'arr')
            matches = np.absolute(zero_row) == np.amax(np.absolute(zero_row))
            col_idx_to_fix = np.where(matches)[0]

            # Break ties for the largest absolute magnitude
            r_idx = np.random.randint(len(col_idx_to_fix))
            col_idx_to_fix = col_idx_to_fix[r_idx]

            # Round the chosen element away from zero
            arr_rounded[zero_row_idx, col_idx_to_fix] = round_away_from_zero(arr[zero_row_idx, col_idx_to_fix])

    # Repeat process for zero columns
    if (~arr_rounded.any(axis=0)).any():

        # Check for zero columns in T_rounded
        zero_col_idxs = np.where(~arr_rounded.any(axis=0))[0]
        for zero_col_idx in zero_col_idxs:
            zero_col = arr[:, zero_col_idx]
            matches = np.absolute(zero_col) == np.amax(np.absolute(zero_col))
            row_idx_to_fix = np.where(matches)[0]

            for i in row_idx_to_fix:
                arr_rounded[i, zero_col_idx] = round_away_from_zero(arr[i, zero_col_idx])
    return arr_rounded.astype(int)

class SupercellTransformation(AbstractTransformation):

    def __init__(self, scaling_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1))):
        """
        Args:
            scaling_matrix: A matrix of transforming the lattice vectors.
                Defaults to the identity matrix. Has to be all integers. e.g.,
                [[2,1,0],[0,3,0],[0,0,1]] generates a new structure with
                lattice vectors a" = 2a + b, b" = 3b, c" = c where a, b, and c
                are the lattice vectors of the original structure.
        """
        self.scaling_matrix = scaling_matrix

    @staticmethod
    def from_scaling_factors(scale_a=1, scale_b=1, scale_c=1):
        """
        Convenience method to get a SupercellTransformation from a simple
        series of three numbers for scaling each lattice vector. Equivalent to
        calling the normal with [[scale_a, 0, 0], [0, scale_b, 0],
        [0, 0, scale_c]]
        Args:
            scale_a: Scaling factor for lattice direction a. Defaults to 1.
            scale_b: Scaling factor for lattice direction b. Defaults to 1.
            scale_c: Scaling factor for lattice direction c. Defaults to 1.
        Returns:
            SupercellTransformation.
        """
        return SupercellTransformation([[scale_a, 0, 0], [0, scale_b, 0], [0, 0, scale_c]])

    def apply_transformation(self, structure, seed=None):
        """
        Apply the transformation.
        Args:
            structure (Structure): Input Structure
        Returns:
            Supercell Structure.
        """
        return structure * self.scaling_matrix

    def __str__(self):
        return "Supercell Transformation with scaling matrix " + "{}".format(self.scaling_matrix)

    def __repr__(self):
        return self.__str__()

    @property
    def inverse(self):
        """
        Raises: NotImplementedError
        """
        raise NotImplementedError()

    @property
    def is_one_to_many(self):
        """
        Returns: False
        """
        return False

class TranslateSitesTransformation(AbstractTransformation):
    """
    This class translates a set of sites by a certain vector.
    """

    def __init__(self, indices_to_move=None, translation_vector=None, vector_in_frac_coords=True):
        """
        Args:
            indices_to_move: The indices of the sites to move
            translation_vector: Vector to move the sites. If a list of list or numpy
                array of shape, (len(indices_to_move), 3), is provided then each
                translation vector is applied to the corresponding site in the
                indices_to_move.
            vector_in_frac_coords: Set to True if the translation vector is in
                fractional coordinates, and False if it is in cartesian
                coordinations. Defaults to True.
        """
        self.indices_to_move = indices_to_move
        self.translation_vector = np.array(translation_vector)
        self.vector_in_frac_coords = vector_in_frac_coords


    def apply_transformation(self, 
        structure, indices_to_move=None, translation_vector=None, 
        vector_in_frac_coords=None, seed=None
    ):
        """
        Apply the transformation.
        Arg:
            structure (Structure): A structurally similar structure in
                regards to crystal and site positions.
        Return:
            Returns a copy of structure with sites translated.
        """
        if(seed is not None):
            np.random.seed(seed)

        if(indices_to_move is None):
            indices_to_move = self.indices_to_move
        if(translation_vector is None):
            translation_vector = self.translation_vector
        else:
            translation_vector = np.array(translation_vector)
        if(vector_in_frac_coords is None):
            vector_in_frac_coords = self.vector_in_frac_coords

        # Get indices to move
        if(indices_to_move is None):
            num_sites = structure.num_sites
            mask_num = max((1, int(np.floor(0.25*num_sites))))
            indices_to_move = np.random.choice(num_sites, mask_num, replace=False)

        if((translation_vector == None).all()):
            translation_vector = np.random.rand(len(indices_to_move),3)

        s = structure.copy()
        if translation_vector.shape == (len(indices_to_move), 3):
            for i, idx in enumerate(indices_to_move):
                s.translate_sites(idx, translation_vector[i], vector_in_frac_coords)
        else:
            s.translate_sites(
                indices_to_move,
                translation_vector,
                vector_in_frac_coords,
            )
        return s

    def __str__(self):
        return "TranslateSitesTransformation for indices " + "{}, vect {} and vect_in_frac_coords = {}".format(
            self.indices_to_move,
            self.translation_vector,
            self.vector_in_frac_coords,
        )

    def __repr__(self):
        return self.__str__()

    @property
    def inverse(self):
        """
        Returns:
            TranslateSitesTranformation with the reverse translation.
        """
        return TranslateSitesTransformation(self.indices_to_move, -self.translation_vector, self.vector_in_frac_coords)

    @property
    def is_one_to_many(self):
        """Return: False"""
        return False

    def as_dict(self):
        """
        Json-serializable dict representation.
        """
        d = MSONable.as_dict(self)
        d["translation_vector"] = self.translation_vector.tolist()
        return d

class CubicSupercellTransformation(AbstractTransformation):
    """
    A transformation that aims to generate a nearly cubic supercell structure
    from a structure.
    The algorithm solves for a transformation matrix that makes the supercell
    cubic. The matrix must have integer entries, so entries are rounded (in such
    a way that forces the matrix to be nonsingular). From the supercell
    resulting from this transformation matrix, vector projections are used to
    determine the side length of the largest cube that can fit inside the
    supercell. The algorithm will iteratively increase the size of the supercell
    until the largest inscribed cube's side length is at least 'min_length'
    and the number of atoms in the supercell falls in the range
    ``min_atoms < n < max_atoms``.
    """

    def __init__(
        self,
        min_atoms: Optional[int] = None,
        max_atoms: Optional[int] = None,
        min_length: float = 15.0,
        force_diagonal: bool = False,
        perturb=None
    ):
        """
        Args:
            max_atoms: Maximum number of atoms allowed in the supercell.
            min_atoms: Minimum number of atoms allowed in the supercell.
            min_length: Minimum length of the smallest supercell lattice vector.
            force_diagonal: If True, return a transformation with a diagonal
                transformation matrix.
            perturb (PerturbStructureTransformation): pre-rotation perturbation
        """
        self.min_atoms = min_atoms if min_atoms else -np.Inf
        self.max_atoms = max_atoms if max_atoms else np.Inf
        self.min_length = min_length
        self.force_diagonal = force_diagonal
        self.transformation_matrix = None
        self.perturb = perturb

    def apply_transformation(self, structure: Structure, seed=None, perturb=None) -> Structure:
        """
        The algorithm solves for a transformation matrix that makes the
        supercell cubic. The matrix must have integer entries, so entries are
        rounded (in such a way that forces the matrix to be nonsingular). From
        the supercell resulting from this transformation matrix, vector
        projections are used to determine the side length of the largest cube
        that can fit inside the supercell. The algorithm will iteratively
        increase the size of the supercell until the largest inscribed cube's
        side length is at least 'num_nn_dists' times the nearest neighbor
        distance and the number of atoms in the supercell falls in the range
        defined by min_atoms and max_atoms.
        Returns:
            supercell: Transformed supercell.
        """

        lat_vecs = structure.lattice.matrix

        # Apply perturbation
        if(perturb is not None):
            self.perturb = perturb
        if(self.perturb is not None):
            structure = self.perturb.apply_transformation(structure)

        # boolean for if a sufficiently large supercell has been created
        sc_not_found = True

        if self.force_diagonal:
            scale = self.min_length / np.array(structure.lattice.abc)
            self.transformation_matrix = np.diag(np.ceil(scale).astype(int))
            st = SupercellTransformation(self.transformation_matrix)
            return st.apply_transformation(structure)

        # target_threshold is used as the desired cubic side lengths
        target_sc_size = self.min_length
        while sc_not_found:
            target_sc_lat_vecs = np.eye(3, 3) * target_sc_size
            self.transformation_matrix = target_sc_lat_vecs @ np.linalg.inv(lat_vecs)

            # round the entries of T and force T to be nonsingular
            self.transformation_matrix = _round_and_make_arr_singular(self.transformation_matrix)  # type: ignore
            if(np.isclose(np.linalg.det(self.transformation_matrix),0)):
                # This fixes the singular matrix issue by simply adding the identity matrix
                self.transformation_matrix += np.eye(3, dtype=int)

            proposed_sc_lat_vecs = self.transformation_matrix @ lat_vecs  # type: ignore

            # Find the shortest dimension length and direction
            a = proposed_sc_lat_vecs[0]
            b = proposed_sc_lat_vecs[1]
            c = proposed_sc_lat_vecs[2]

            length1_vec = c - _proj(c, a)  # a-c plane
            length2_vec = a - _proj(a, c)
            length3_vec = b - _proj(b, a)  # b-a plane
            length4_vec = a - _proj(a, b)
            length5_vec = b - _proj(b, c)  # b-c plane
            length6_vec = c - _proj(c, b)
            length_vecs = np.array(
                [
                    length1_vec,
                    length2_vec,
                    length3_vec,
                    length4_vec,
                    length5_vec,
                    length6_vec,
                ]
            )

            # Get number of atoms
            st = SupercellTransformation(self.transformation_matrix)
            superstructure = st.apply_transformation(structure)
            num_at = superstructure.num_sites

            # Check if constraints are satisfied
            if (
                np.min(np.linalg.norm(length_vecs, axis=1)) >= self.min_length
                and self.min_atoms <= num_at <= self.max_atoms
            ):
                return superstructure
            # Increase threshold until proposed supercell meets requirements
            target_sc_size += 0.1
            if num_at > self.max_atoms:
                raise AttributeError(
                    "While trying to solve for the supercell, the max "
                    "number of atoms was exceeded. Try lowering the number"
                    "of nearest neighbor distances."
                )
        raise AttributeError("Unable to find cubic supercell")

    @property
    def inverse(self):
        """
        Returns:
            None
        """
        return None

    @property
    def is_one_to_many(self):
        """
        Returns:
            False
        """
        return False

class PrimitiveCellTransformation(AbstractTransformation):
    """
    This class finds the primitive cell of the input structure.
    It returns a structure that is not necessarily orthogonalized
    Author: Will Richards
    """

    def __init__(self, tolerance=0.5):
        """
        Args:
            tolerance (float): Tolerance for each coordinate of a particular
                site. For example, [0.5, 0, 0.5] in cartesian coordinates will be
                considered to be on the same coordinates as [0, 0, 0] for a
                tolerance of 0.5. Defaults to 0.5.
        """
        self.tolerance = tolerance

    def apply_transformation(self, structure, seed=None):
        """
        Returns most primitive cell for structure.
        Args:
            structure: A structure
        Returns:
            The most primitive structure found. The returned structure is
            guaranteed to have len(new structure) <= len(structure).
        """
        return structure.get_primitive_structure(tolerance=self.tolerance)

    def __str__(self):
        return "Primitive cell transformation"

    def __repr__(self):
        return self.__str__()

    @property
    def inverse(self):
        """
        Returns: None
        """
        return None

    @property
    def is_one_to_many(self):
        """
        Returns: False
        """
        return False




import torch
from torch import nn
import torch.nn.functional as F
from torch.nn import Linear, LayerNorm, ReLU

from torch_scatter import scatter
from torch_geometric.nn import MessagePassing
from torch_geometric.utils import add_self_loops, degree, softmax
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool

from auglichem.utils import NUM_ATOM_TYPE, NUM_CHIRALITY_TAG, NUM_BOND_TYPE, NUM_BOND_DIRECTION



class GINEConv(MessagePassing):
    def __init__(self, emb_dim, aggr="add"):
        super(GINEConv, self).__init__()
        self.mlp = nn.Sequential(
            nn.Linear(emb_dim, 2*emb_dim), 
            nn.ReLU(), 
            nn.Linear(2*emb_dim, emb_dim)
        )
        self.edge_embedding1 = nn.Embedding(NUM_BOND_TYPE, emb_dim)
        self.edge_embedding2 = nn.Embedding(NUM_BOND_DIRECTION, emb_dim)

        nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        nn.init.xavier_uniform_(self.edge_embedding2.weight.data)
        self.aggr = aggr

    def forward(self, x, edge_index, edge_attr):
        # add self loops in the edge space
        edge_index = add_self_loops(edge_index, num_nodes=x.size(0))[0]
        # print(edge_index)

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:,0] = 4 #bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        edge_embeddings = self.edge_embedding1(edge_attr[:,0]) + self.edge_embedding2(edge_attr[:,1])

        # return self.propagate(self.aggr, edge_index, x=x, edge_attr=edge_embeddings)
        return self.propagate(edge_index, x=x, edge_attr=edge_embeddings)

    def message(self, x_j, edge_attr):
        return x_j + edge_attr

    def update(self, aggr_out):
        return self.mlp(aggr_out)


class GINE(nn.Module):
    """
    Args:
        num_layer (int): the number of GNN layers
        emb_dim (int): dimensionality of embeddings
        JK (str): last, concat, max or sum.
        max_pool_layer (int): the layer from which we use max pool rather than add pool for neighbor aggregation
        drop_ratio (float): dropout rate
        gnn_type: gin, gcn, graphsage, gat
    Output:
        node representations
    """
    def __init__(self, 
        task='classification', emb_dim=300, 
        feat_dim=256, num_layer=5, pool='mean', drop_ratio=0, output_dim=None, **kwargs
    ):
        super(GINE, self).__init__()
        self.num_layer = num_layer
        self.emb_dim = emb_dim
        self.feat_dim = feat_dim
        self.drop_ratio = drop_ratio
        self.task = task
        if(output_dim == None):
            self.output_dim = 1
        else:
            self.output_dim = output_dim

        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.x_embedding1 = nn.Embedding(NUM_ATOM_TYPE, emb_dim)
        self.x_embedding2 = nn.Embedding(NUM_CHIRALITY_TAG, emb_dim)

        nn.init.xavier_uniform_(self.x_embedding1.weight.data)
        nn.init.xavier_uniform_(self.x_embedding2.weight.data)

        # List of MLPs
        self.gnns = nn.ModuleList()
        for layer in range(num_layer):
            self.gnns.append(GINEConv(emb_dim, aggr="add"))

        # List of batchnorms
        self.batch_norms = nn.ModuleList()
        for layer in range(num_layer):
            self.batch_norms.append(nn.BatchNorm1d(emb_dim))

        if pool == 'mean':
            self.pool = global_mean_pool
        elif pool == 'add':
            self.pool = global_add_pool
        elif pool == 'max':
            self.pool = global_max_pool
        else:
            raise ValueError('Not defined pooling!')
        
        self.feat_lin = nn.Linear(self.emb_dim, self.feat_dim)

        if self.task == 'classification':
            self.output_dim *= 2
            self.pred_lin = nn.Sequential(
                nn.Linear(self.feat_dim, self.feat_dim//2), 
                nn.ReLU(inplace=True),
                nn.Linear(self.feat_dim//2, self.output_dim)
            )
        elif self.task == 'regression':
            self.pred_lin = nn.Sequential(
                nn.Linear(self.feat_dim, self.feat_dim//2), 
                nn.ReLU(inplace=True),
                nn.Linear(self.feat_dim//2, self.feat_dim//2),
                nn.ReLU(inplace=True),
                nn.Linear(self.feat_dim//2, self.feat_dim//4),
                nn.ReLU(inplace=True),
                nn.Linear(self.feat_dim//4, self.output_dim)
            )

    def forward(self, data):
        x = data.x
        edge_index = data.edge_index
        edge_attr = data.edge_attr

        # x = self.x_embedding1(x[:,0]) + self.x_embedding2(x[:,1])

        # h_list = [x]
        # for layer in range(self.num_layer):
        #     h = self.gnns[layer](h_list[layer], edge_index, edge_attr)
        #     h = self.batch_norms[layer](h)
        #     if layer == self.num_layer - 1:
        #         h = F.dropout(h, self.drop_ratio, training = self.training)
        #     else:
        #         h = F.dropout(F.relu(h), self.drop_ratio, training = self.training)
        #     h_list.append(h)
        
        h = self.x_embedding1(x[:,0]) + self.x_embedding2(x[:,1])

        for layer in range(self.num_layer):
            h = self.gnns[layer](h, edge_index, edge_attr)
            h = self.batch_norms[layer](h)
            # if layer == self.num_layer - 1:
            #     h = F.dropout(h, self.drop_ratio, training=self.training)
            # else:
            #     h = F.dropout(F.relu(h), self.drop_ratio, training=self.training)
            h = F.dropout(F.relu(h), self.drop_ratio, training=self.training)

        h = self.feat_lin(h)
        h = self.pool(h, data.batch)

        return h, self.pred_lin(h)

    def load_my_state_dict(self, state_dict):
        own_state = self.state_dict()
        for name, param in state_dict.items():
            if name not in own_state:
                continue
            if isinstance(param, nn.parameter.Parameter):
                # backwards compatibility for serialized parameters
                param = param.data
            own_state[name].copy_(param)


if __name__ == "__main__":
    model = GINet()
    print(model)
    layer_list = []
    for name, param in model.named_parameters():
        if 'pred_lin' in name:
            print(name, param.requires_grad)
            layer_list.append(name)

    params = list(map(lambda x: x[1],list(filter(lambda kv: kv[0] in layer_list, model.named_parameters()))))
    base_params = list(map(lambda x: x[1],list(filter(lambda kv: kv[0] not in layer_list, model.named_parameters()))))
    print(base_params)
    print(params)

    from torch.optim import Adam
    optimizer = Adam([{'params': base_params, 'lr': 1e-4}, {'params': params}], lr=1e-3)



#import auglichem.utils._splitting as splitting
from auglichem.utils._splitting import scaffold_split, random_split
from auglichem.utils._constants import (
        ATOM_LIST,
        NUM_ATOM_TYPE,
        CHIRALITY_LIST,
        NUM_CHIRALITY_TAG,
        BOND_LIST,
        NUM_BOND_TYPE,
        BONDDIR_LIST,
        NUM_BOND_DIRECTION
)

__all__ = [
        "ATOM_LIST",
        "NUM_ATOM_TYPE",
        "CHIRALITY_LIST",
        "NUM_CHIRALITY_TAG",
        "BOND_LIST",
        "NUM_BOND_TYPE",
        "BONDDIR_LIST",
        "NUM_BOND_DIRECTION",
        "scaffold_split",
        "random_split"
]



import torch
import scipy.spatial

#if torch.cuda.is_available():
#    import torch_cluster.radius_cuda


def radius(x, y, r, batch_x=None, batch_y=None, max_num_neighbors=32):
    r"""Finds for each element in :obj:`y` all points in :obj:`x` within
    distance :obj:`r`.

    Args:
        x (Tensor): Node feature matrix
            :math:`\mathbf{X} \in \mathbb{R}^{N \times F}`.
        y (Tensor): Node feature matrix
            :math:`\mathbf{Y} \in \mathbb{R}^{M \times F}`.
        r (float): The radius.
        batch_x (LongTensor, optional): Batch vector
            :math:`\mathbf{b} \in {\{ 0, \ldots, B-1\}}^N`, which assigns each
            node to a specific example. (default: :obj:`None`)
        batch_y (LongTensor, optional): Batch vector
            :math:`\mathbf{b} \in {\{ 0, \ldots, B-1\}}^M`, which assigns each
            node to a specific example. (default: :obj:`None`)
        max_num_neighbors (int, optional): The maximum number of neighbors to
            return for each element in :obj:`y`. (default: :obj:`32`)

    :rtype: :class:`LongTensor`

    .. testsetup::

        import torch
        from torch_cluster import radius

    .. testcode::


        >>> x = torch.Tensor([[-1, -1], [-1, 1], [1, -1], [1, 1]])
        >>> batch_x = torch.tensor([0, 0, 0, 0])
        >>> y = torch.Tensor([[-1, 0], [1, 0]])
        >>> batch_y = torch.tensor([0, 0])
        >>> assign_index = radius(x, y, 1.5, batch_x, batch_y)
    """

    if batch_x is None:
        batch_x = x.new_zeros(x.size(0), dtype=torch.long)

    if batch_y is None:
        batch_y = y.new_zeros(y.size(0), dtype=torch.long)

    x = x.view(-1, 1) if x.dim() == 1 else x
    y = y.view(-1, 1) if y.dim() == 1 else y

    assert x.dim() == 2 and batch_x.dim() == 1
    assert y.dim() == 2 and batch_y.dim() == 1
    assert x.size(1) == y.size(1)
    assert x.size(0) == batch_x.size(0)
    assert y.size(0) == batch_y.size(0)

    # Need to put data on CPU for this function, hold on to device to put back
    _device = x.device

    x = torch.cat([x, 2 * r * batch_x.view(-1, 1).to(x.dtype)], dim=-1)
    y = torch.cat([y, 2 * r * batch_y.view(-1, 1).to(y.dtype)], dim=-1)

    tree = scipy.spatial.cKDTree(x.cpu().detach().numpy())
    _, col = tree.query(
        y.cpu().detach().numpy(), k=max_num_neighbors, distance_upper_bound=r + 1e-8)
    col = [torch.from_numpy(c).to(torch.long) for c in col]
    row = [torch.full_like(c, i) for i, c in enumerate(col)]
    row, col = torch.cat(row, dim=0), torch.cat(col, dim=0)
    mask = col < int(tree.n)
    return torch.stack([row[mask], col[mask]], dim=0).to(device=_device)



def radius_graph(x,
                 r,
                 batch=None,
                 loop=False,
                 max_num_neighbors=32,
                 flow='source_to_target'):
    r"""Computes graph edges to all points within a given distance.

    Args:
        x (Tensor): Node feature matrix
            :math:`\mathbf{X} \in \mathbb{R}^{N \times F}`.
        r (float): The radius.
        batch (LongTensor, optional): Batch vector
            :math:`\mathbf{b} \in {\{ 0, \ldots, B-1\}}^N`, which assigns each
            node to a specific example. (default: :obj:`None`)
        loop (bool, optional): If :obj:`True`, the graph will contain
            self-loops. (default: :obj:`False`)
        max_num_neighbors (int, optional): The maximum number of neighbors to
            return for each element in :obj:`y`. (default: :obj:`32`)
        flow (string, optional): The flow direction when using in combination
            with message passing (:obj:`"source_to_target"` or
            :obj:`"target_to_source"`). (default: :obj:`"source_to_target"`)

    :rtype: :class:`LongTensor`

    .. testsetup::

        import torch
        from torch_cluster import radius_graph

    .. testcode::

        >>> x = torch.Tensor([[-1, -1], [-1, 1], [1, -1], [1, 1]])
        >>> batch = torch.tensor([0, 0, 0, 0])
        >>> edge_index = radius_graph(x, r=1.5, batch=batch, loop=False)
    """

    assert flow in ['source_to_target', 'target_to_source']
    row, col = radius(x, x, r, batch, batch, max_num_neighbors + 1)
    row, col = (col, row) if flow == 'source_to_target' else (row, col)
    if not loop:
        mask = row != col
        row, col = row[mask], col[mask]
    return torch.stack([row, col], dim=0)



import torch
from torch import nn
import torch.nn.functional as F
from torch.nn import Linear, LayerNorm, ReLU

from torch_scatter import scatter
# from torch_geometric.nn import GINEConv
from torch_geometric.nn import MessagePassing
from torch_geometric.utils import add_self_loops, degree, softmax
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool


ATOM_LIST = list(range(1,100))
num_atom_type = len(ATOM_LIST) + 1 # including the extra mask tokens

num_bond_type = 2 # self-connected and otherwise


class GINEConv(MessagePassing):
    def __init__(self, emb_dim, aggr="add"):
        super(GINEConv, self).__init__()
        self.mlp = nn.Sequential(
            nn.Linear(emb_dim, 2*emb_dim), 
            nn.Softplus(), 
            nn.Linear(2*emb_dim, emb_dim)
        )
        self.edge_embedding1 = nn.Embedding(num_bond_type, emb_dim)
        # self.edge_embedding2 = nn.Embedding(num_bond_direction, emb_dim)

        nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        # nn.init.xavier_uniform_(self.edge_embedding2.weight.data)
        self.aggr = aggr

    def forward(self, x, edge_index, edge_attr):
        # add self loops in the edge space
        edge_index = add_self_loops(edge_index, num_nodes=x.size(0))[0]

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.ones(x.size(0)) * (num_bond_type-1) # bond type for self-loop edge
        # self_loop_attr[:,0] = 4
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        # edge_embeddings = self.edge_embedding1(edge_attr[:,0]) + self.edge_embedding2(edge_attr[:,1])
        edge_embeddings = self.edge_embedding1(edge_attr)

        # return self.propagate(self.aggr, edge_index, x=x, edge_attr=edge_embeddings)
        return self.propagate(edge_index, x=x, edge_attr=edge_embeddings)

    def message(self, x_j, edge_attr):
        return x_j + edge_attr

    def update(self, aggr_out):
        return self.mlp(aggr_out)


class GINet(nn.Module):
    """
    Args:
        num_layer (int): the number of GNN layers
        emb_dim (int): dimensionality of embeddings
        feat_dim (int): feature dimenstion
        drop_ratio (float): dropout rate
        pool (str): pooling function
    Output:
        node representations
    """
    def __init__(self, num_layer=5, emb_dim=256, feat_dim=512, drop_ratio=0, pool='mean'):
        super(GINet, self).__init__()
        self.num_layer = num_layer
        self.emb_dim = emb_dim
        self.feat_dim = feat_dim
        self.num_atom_type = num_atom_type
        self.drop_ratio = drop_ratio

        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.x_embedding1 = nn.Embedding(num_atom_type, emb_dim)
        self.x_embedding2 = nn.Linear(3, emb_dim)

        nn.init.xavier_uniform_(self.x_embedding1.weight.data)
        # nn.init.xavier_uniform_(self.x_embedding2.weight.data)

        # List of MLPs
        self.gnns = nn.ModuleList()
        for layer in range(num_layer):
            self.gnns.append(GINEConv(emb_dim, aggr="add"))

        # List of batchnorms
        self.batch_norms = nn.ModuleList()
        for layer in range(num_layer):
            self.batch_norms.append(nn.BatchNorm1d(emb_dim))

        self.feat_lin = nn.Linear(self.emb_dim, self.feat_dim)

        if pool == 'mean':
            self.pool = global_mean_pool
        elif pool == 'max':
            self.pool = global_max_pool
        elif pool == 'add':
            self.pool = global_add_pool
        else:
            raise ValueError('Pooling must be either mean, max, or add')

        self.head = nn.Sequential(
            nn.Linear(self.feat_dim, self.feat_dim//2), 
            nn.Softplus(),
            nn.Linear(self.feat_dim//2, 1)
        )

    def forward(self, data):
        edge_index = data.edge_index
        edge_attr = data.edge_attr
        
        h = self.x_embedding1(data.atomics) + self.x_embedding2(data.pos)
        # h = self.x_embedding1(data.atomics) + \
        #     self.x_embedding2(data.pos) + \
        #     self.x_embedding3(data.feat)

        for layer in range(self.num_layer):
            h = self.gnns[layer](h, edge_index, edge_attr)
            h = self.batch_norms[layer](h)
            if layer == self.num_layer - 1:
                h = F.dropout(h, self.drop_ratio, training=self.training)
            else:
                h = F.dropout(F.softplus(h), self.drop_ratio, training=self.training)

        feat = self.feat_lin(h)
        feat = self.pool(feat, data.batch)
        out = self.head(feat)

        self.feat = feat
        
        return out


if __name__ == "__main__":
    model = GINet()
    print(model)



import os
import urllib
from zipfile import ZipFile
from tqdm import tqdm
from distutils.dir_util import copy_tree
import shutil

import requests
import zipfile

import numpy as np
import time
import csv
import json

from torch.utils.data.dataloader import default_collate

USER_AGENT = "auglichem"

#TODO: Automated data loading once we host the datasets

# Data available at:
#        https://drive.google.com/drive/folders/1R8kjl1O1cGn-bzKij4IvzCrzLOUt0139?usp=sharing


class AtomInitializer(object):
    """
    Base class for intializing the vector representation for atoms.

    !!! Use one AtomInitializer per dataset !!!
    """
    def __init__(self, atom_types):
        self.atom_types = set(atom_types)
        self._embedding = {}

    def get_atom_feat(self, atom_type):
        assert atom_type in self.atom_types
        return self._embedding[atom_type]

    def get_atom_features(self, atomics):
        return np.vstack([self.get_atom_feat(i) for i in atomics])

    def load_state_dict(self, state_dict):
        self._embedding = state_dict
        self.atom_types = set(self._embedding.keys())
        self._decodedict = {idx: atom_type for atom_type, idx in
                            self._embedding.items()}

    def state_dict(self):
        return self._embedding

    def decode(self, idx):
        if not hasattr(self, '_decodedict'):
            self._decodedict = {idx: atom_type for atom_type, idx in
                                self._embedding.items()}
        return self._decodedict[idx]

class AtomCustomJSONInitializer(AtomInitializer):
    """
    Initialize atom feature vectors using a JSON file, which is a python
    dictionary mapping from element number to a list representing the
    feature vector of the element.

    Parameters
    ----------

    elem_embedding_file: str
        The path to the .json file
    """
    def __init__(self, elem_embedding_file):
        with open(elem_embedding_file) as f:
            elem_embedding = json.load(f)
        f.close()
        elem_embedding = {int(key): value for key, value
                          in elem_embedding.items()}
        atom_types = set(elem_embedding.keys())
        super(AtomCustomJSONInitializer, self).__init__(atom_types)
        for key, value in elem_embedding.items():
            self._embedding[key] = np.array(value, dtype=float)


# Download functions taken from ManufactureNet
def download_file_from_google_drive(id, destination):
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id }, stream = True)
    token = get_confirm_token(response)

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    save_response_content(response, destination)

def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None


def save_response_content(response, destination):
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in tqdm(response.iter_content(CHUNK_SIZE)):
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)


def extract_files(name, root):
    Zip = ZipFile(name)
    Zip.extractall(root)


def remove_zip(name):
    os.remove(name)


def download_url(url, root, filename=None):
    """Download a file from a url and place it in root.
    Args:
        url (str): URL to download file from
        root (str): Directory to place downloaded file in
        filename (str, optional): Name to save the file under. If None, use the basename of
                                  the URL
    """
    URL = "https://docs.google.com/uc?export=download/"
    destination = root+filename
    if(os.path.isdir(destination[:-4])):
        print("Data found at: {}".format(destination[:-4]))
        return destination[:-4]
    print("Downloading data to: {}...".format(destination[:-4]))
    download_file_from_google_drive(url, destination)
    print("Extracting zipfile...")
    extract_files(destination, root)
    print("Removing zipfile...")
    remove_zip(destination)

    return destination[:-4]

def _load_data(dataset, data_path='./data_download', data_src=None):
    '''
        Loads dataset, sets task to regression for the downloadable sets, gets the atom
        embedding file, and updates the data path.
    '''
    # Data hosting is done with google drive. You must create a publicly available link
    # using "get link", then convert it to a downloadable link here:
    # https://sites.google.com/site/gdocs2direct/
    if(dataset == 'lanthanides'):
        task = 'regression'
        target = ["formation_energy"] #TODO: Need to verify
        data_id = "1AB0XkOUdmOkxfKKjW_XkAxxGhAuA4F5k"
        csv_file_path = download_url(data_id, data_path, "/lanths.zip")
        csv_file_path = data_path + "/lanths/id_prop.csv"
        embedding_path = data_path + "/lanths/atom_init.json"
        data_path += "/lanths"
    elif(dataset == 'band_gap'):
        task = 'regression'
        target = ["band_gap"]
        data_id = "1pr8yxK3-BdfQwiVG_V8s_eCITY2l_ynC"
        csv_file_path = download_url(data_id, data_path, "/band.zip")
        csv_file_path = data_path + "/band/id_prop.csv"
        embedding_path = data_path + "/band/atom_init.json"
        data_path += "/band"
    elif(dataset == 'perovskites'):
        task = 'regression'
        target = ["energy"] #TODO: Need to verify
        data_id = "1y-SAMNBHrE9GUCuZofDuvLXc3MXGF7dG"
        csv_file_path = download_url(data_id, data_path, "/abx3_cifs.zip")
        csv_file_path = data_path + "/abx3_cifs/id_prop.csv"
        embedding_path = data_path + "/abx3_cifs/atom_init.json"
        data_path += "/abx3_cifs"
    elif(dataset == 'fermi_energy'):
        task = 'regression'
        target = ["fermi_energy"]
        data_id = "1m5I2MSfoT9WOlua4H4ZBYQS3dEzpmYid"
        csv_file_path = download_url(data_id, data_path, "/fermi.zip")
        csv_file_path = data_path + "/fermi/id_prop.csv"
        embedding_path = data_path + "/fermi/atom_init.json"
        data_path += "/fermi"
    elif(dataset == 'formation_energy'):
        task = 'regression'
        target = ["formation_energy"]
        data_id = "1NNmald3DfhHUr2gzN4V1ExsTp2cuYgvx"
        csv_file_path = download_url(data_id, data_path, "/FE.zip")
        csv_file_path = data_path + "/FE/id_prop.csv"
        embedding_path = data_path + "/FE/atom_init.json"
        data_path += "/FE"
    elif(dataset == "HOIP"):
        task = 'regression'
        target = ["gvrh"]
        data_id = "1hQehDWxaBO5K74aOg5O1RDVwIjmeQLvo"
        csv_file_path = download_url(data_id, data_path, "/HOIP.zip")
        csv_file_path = data_path + "/HOIP/id_prop.csv"
        embedding_path = data_path + "/HOIP/atom_init.json"
        data_path += "/HOIP"
    elif(dataset == "custom"):
        task = 'regression'
        target = ["target"]
        data_path += "/" + data_src.split("/")[-1]

        # Check if files are already there
        if(os.path.isdir(data_path)):
            print("Data directory found at: {}".format(data_path))
        else:
            shutil.copytree(data_src, data_path) # Copy from src

        # Check if atom_init.json has been downloaded
        if(os.path.exists(data_path+"/atom_init.json")):
            print("atom_init.json found at: {}".format(data_path))
        else:
            # Download atom_init.json
            data_id = "13vYFP-MIlYAsvuvfdjOpkUn9UAXKFtuB"
            csv_file_path = download_file_from_google_drive(data_id, data_path+"/atom_init.json")

        # Updated relevant paths
        csv_file_path = data_path + "/id_prop.csv"
        embedding_path = data_path + "/atom_init.json"
    else:
        raise ValueError("Please select one of the following datasets: lanthanides, band_gap, perovskites, fermi_energy, formation_energy, GVRH, HOIP, is_metal")
        

    return data_path, embedding_path, csv_file_path, target, task


def read_crystal(dataset, data_path, data_src=None):
    """
        Inputs:
        -------
        dataset (str): The dataset to be used, needs to be one of: lanthanides, band_gap,
                       perovskites, fermi_energy, or formation_energy
        data_path (str): The path to search for data. If the requested data set is not there,
                         the data is downloaded automatically and stored at data_path.
        data_src (str, default=None): Source directory for using custom data set.

        Outputs:
        --------
        csv_file_path (str): The path of id_prop.csv in the data set directory.
        embedding_file_path (str): The path of [] in the data set directory.
        ari: (AtomCustomJSONInitializer): Object that initializes the vector representation
                                          of atoms.
        data_path (str): The relative path of the data.
        target (str): The name of the target variable.
        task (str): regression or classification. Currently only classification is supported.
    """

    # Create data download directory if it does not exist
    if(data_path is None):
        data_path = './data_download'
    if(not os.path.exists(data_path)):
        os.mkdir(data_path)

    # Download files if not already there
    data_path, embedding_path, csv_file_path, target, task = _load_data(dataset, data_path,
                                                                        data_src)
    return csv_file_path, \
           embedding_path, \
           AtomCustomJSONInitializer(embedding_path), \
           data_path, \
           target, \
           task

